import dataclasses
import enum

from p2g import axis
from p2g import coords
from p2g import err
from p2g import gbl
from p2g import nd
from p2g import stat


# fast and slow and neither look up params.
# can be changed by using explicit feed.
# kwargs could contain x,y,z,feed ec
# as could extras


class MovementSpace(enum.IntEnum):
    WORK = enum.auto()
    MACHINE = enum.auto()
    R9810 = enum.auto()


class MovementRelOrAbs(enum.IntEnum):
    RELATIVE = enum.auto()
    ABSOLUTE = enum.auto()


class MovementOrder(enum.IntEnum):
    XYZ = enum.auto()
    XY_THEN_Z = enum.auto()
    Z_THEN_XY = enum.auto()


def do_goto_worker(self, fter, args, kwargs):
    # split out arguments we understand from
    # ones for coordinates.

    def accumulate_relorabs():
        match self.relorabs_:
            case MovementRelOrAbs.ABSOLUTE:
                yield "G90"

            case MovementRelOrAbs.RELATIVE:
                yield "G91"

    def accumulate_probe():
        if self.probe_:
            yield "G31"

    def accumulate_mcode():
        if self.mcode_:
            yield self.mcode_

    def accumulate_feed():
        if self.feed_ == 0.0:
            err.compiler("Need feed rate.")

        yield f"F{nd.to_gcode(self.feed_)}"

    def accumulate_parts():
        match self.space_:
            case MovementSpace.MACHINE:
                yield "G01"
                yield "G53"
                yield from accumulate_relorabs()
                yield from accumulate_probe()
                yield from accumulate_mcode()
            case MovementSpace.WORK:
                yield "G01"
                yield from accumulate_relorabs()
                yield from accumulate_probe()
                yield from accumulate_mcode()
            case MovementSpace.R9810:
                yield "G65 R9810"
                if self.relorabs_ & MovementRelOrAbs.RELATIVE:
                    err.compiler("Relative with 9810 move is illegal.")

                if self.probe_:
                    err.compiler("Probe with 9810 move is illegal.")

                if self.mcode_:
                    err.compiler("MCODE with 9810 move is illegal.")

        yield from accumulate_feed()

    def get_coords():
        values = coords.unpack(args, kwargs)
        for aname, value in zip(axis.low_names_v(), values):
            # skip non important coords
            if fter and aname not in fter:
                continue
            # skip non mentioned coords
            if value.is_none_constant:
                continue
            yield aname, value

    def accumulate_coords(cos):
        for aname, value in cos:
            yield f"{aname}{value.to_gcode(nd.NodeModifier.EMPTY)}"

    cos = list(get_coords())
    if not cos:
        return

    stat.code(
        accumulate_parts(),
        accumulate_coords(cos),
    )


@dataclasses.dataclass(eq=True, frozen=True)
class GotoWorker:
    user_defined = True
    want_bp_: bool
    space_: MovementSpace
    order_: MovementOrder
    relorabs_: MovementRelOrAbs
    feed_: float
    probe_: bool
    mcode_: str

    def to_symtab_entry(self, *_) -> str:
        return "".join(
            [
                str(self.feed_),
                " ",
                self.mcode_ + " " if self.mcode_ else "",
                self.space_.name.lower(),
                " ",
                self.order_.name.lower(),
                " ",
                "probe " if self.probe_ else "",
                "bp" if self.want_bp_ else "",
            ]
        )

    def update(self, key, value):
        xxx = {**self.__dict__, key: value}
        return GotoWorker(**xxx)

    @property
    def work(self):
        return self.update("space_", MovementSpace.WORK)

    @property
    def r9810(self):
        return self.update("space_", MovementSpace.R9810)

    @property
    def machine(self):
        return self.update("space_", MovementSpace.MACHINE)

    @property
    def z_then_xy(self):
        return self.update("order_", MovementOrder.Z_THEN_XY)

    @property
    def xy_then_z(self):
        return self.update("order_", MovementOrder.XY_THEN_Z)

    @property
    def probe(self):
        return self.update("probe_", True)

    @property
    def xyz(self):
        return self.update("order_", MovementOrder.XYZ)

    @property
    def relative(self):
        return self.update("relorabs_", MovementRelOrAbs.RELATIVE)

    @property
    def absolute(self):
        return self.update("relorabs_", MovementRelOrAbs.ABSOLUTE)

    def feed(self, feed):
        return self.update("feed_", feed)

    def mcode(self, m_code):
        return self.update("mcode_", str(m_code))

    def __call__(self, *args, **kwargs):
        match self.order_:
            case MovementOrder.XYZ:
                do_goto_worker(self, "", args, kwargs)
            case MovementOrder.Z_THEN_XY:
                do_goto_worker(self, "z", args, kwargs)
                do_goto_worker(self, "xy", args, kwargs)
            case MovementOrder.XY_THEN_Z:
                do_goto_worker(self, "xy", args, kwargs)
                do_goto_worker(self, "z", args, kwargs)


gbl.HasToSymTab.register(GotoWorker)
goto = GotoWorker(
    want_bp_=False,
    space_=MovementSpace.WORK,
    relorabs_=MovementRelOrAbs.ABSOLUTE,
    order_=MovementOrder.XYZ,
    feed_=0.0,
    probe_=False,
    mcode_="",
)



import pathlib
import re

import rich.console

from rich.table import Table

from p2g import lib


# three sorts of py outputs, as an arrays,  as attributes to a
# class, or straight into global names.

INTO_ARRAY = 0
INDENT = "    "
INTO_ATTR = 0
INDENT = "    "

INTO_GBL = 1
INDENT = ""


def def_prefix(key):
    if INTO_ARRAY:  # no cover
        return f'dst["{key}"]'
    if INTO_ATTR:  # no cover
        return f"dst.{key}"

    return f"{key}"


MAKE_PFX = "p2g."


# pylint: disable=invalid-name
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-statements
class NAXES:
    def __init__(self):
        pass

    def __str__(self):
        return "NAXES"


class MacroVar:
    def __init__(self, *, key, addr, alias=None, size=None, name="", typ=""):
        self.typ = typ
        self.key = key
        self.addr = addr

        self.name = name
        self.alias = alias
        self.size = size
        suffix = ""
        self.last = None
        if isinstance(self.size, NAXES):
            self.size_as_text = "naxes"
            self.last = NAXES
            suffix = "…"
        else:
            self.last = self.addr + self.size - 1
            # that's a non breaking space
            self.size_as_text = f"{self.size:5}"

            if self.last != self.addr:
                suffix = f" … #{self.last:5}"
        self.range_as_text = f"#{self.addr:5}{suffix}".center(15)

    def for_regen(self):
        funcname = self.__class__.__qualname__
        res = [str(self.addr)]
        if self.size is not None:
            res.append(f"size={self.size}")
        restxt = ", ".join(res)
        return f"{def_prefix(self.name)} = {funcname}({restxt})"

    # for the .txt output
    def for_txt(self):
        return (
            self.range_as_text,
            str(self.size),
            self.key,
            self.typ,
            self.name,
        )

    def for_org(self):
        return (
            f"| <code>{self.name}</code> "
            + f"| <code>{self.size_as_text}</code>"
            + f"| <code>{self.range_as_text}</code> |\n"
        )

    def for_py_out(self):
        return f"# {self.addr} .. {self.last} {self.name} .."

    def __lt__(self, other):
        return self.addr < other.addr


class Gen(MacroVar):
    def __init__(
        self,
        *,
        exp_name,
        idc,
        addr,
        size=None,
        typ,
    ):
        self.exp_name = exp_name

        super().__init__(
            key=idc,
            addr=addr,
            size=size,
            typ=typ,
        )

    def for_py_out(self):
        rs = ""

        sstr = ""
        if self.size is not None and self.size != 1:
            sstr = f"[{self.size}]"
        return f"{def_prefix(self.name)} = {MAKE_PFX}Fixed{sstr}({rs}addr={self.addr})"


def one(addr, typ="Float"):
    return Gen(
        exp_name="Fixed",
        idc="v" + typ[0],
        addr=addr,
        size=1,
        typ=typ,
    )


def ione(addr):
    return one(addr, typ="Int")


class Alias(MacroVar):
    def __init__(self, src):
        super().__init__(key="A", addr=src.addr, size=src.size, alias=src)

    def for_regen(self):
        assert self.alias is not None
        return f"{def_prefix(self.name.ljust(20))} = alias({def_prefix}{self.alias.name})"

    def for_py_out(self):
        assert self.alias is not None
        return f"{def_prefix(self.name)} = {MAKE_PFX}alias({def_prefix(self.alias.name)})"

    def for_txt(self):
        return (
            self.range_as_text,
            str(self.size),
            self.key,
            self.typ,
            " also " + self.name,
        )


def fixed(addr, size, typ="Float"):
    return Gen(
        exp_name="Fixed",
        idc="V",
        addr=addr,
        size=size,
        typ=typ,
    )


class Gap(MacroVar):
    C = 0

    def __init__(self, addr, size=None):
        super().__init__(key="_", addr=addr, size=size)
        self.name = "-"

    def for_regen(self):
        self.name = f"gap{Gap.C:02}"
        Gap.C += 1
        return super().for_regen()


def cwcpos(addr):
    return Gen(exp_name="CWCPos", idc="m", addr=addr, size=NAXES(), typ="Float")


def machinepos(addr):
    return Gen(exp_name="MachinePos", idc="m", addr=addr, size=NAXES(), typ="Float")


def workoffsettable(addr):
    return Gen(exp_name="WorkOffsetTable", idc="W", addr=addr, typ="Float", size=NAXES())


def tooltable(addr, size, typ="Float"):
    return Gen(exp_name="ToolTable", idc="T", addr=addr, size=size, typ=typ)


class PalletTable(MacroVar):
    def __init__(self, addr, size):
        super().__init__(key="L", addr=addr, size=size, typ="Int")


class Names:
    interesting: list[MacroVar]

    def __init__(self):
        object.__setattr__(self, "interesting", [])

    def py_write_to(self, out):
        for v in self.interesting:
            out.write(INDENT + v.for_py_out() + "\n")

    def __setattr__(self, key, value):
        if key[0].isupper():
            value.name = key
            object.__setattr__(self, key, value)

        self.interesting.append(value)


class HaasNames(Names):
    title = "HAAS"

    def __init__(self):
        super().__init__()

        self.NULL = one(0)
        self.MACRO_ARGUMENTS = fixed(1, size=33)
        self.gap01 = Gap(34, size=66)
        self.GP_SAVED1 = fixed(100, size=100)
        self.gap02 = Gap(200, size=300)
        self.GP_SAVED2 = fixed(500, size=50)

        self.PROBE_CALIBRATION1 = fixed(550, size=6)
        self.PROBE_R = fixed(556, size=3)
        self.PROBE_CALIBRATION2 = fixed(559, size=22)

        self.GP_SAVED3 = fixed(581, size=119)
        self.gap03 = Gap(700, size=100)
        self.GP_SAVED4 = fixed(800, size=200)
        self.INPUTS = fixed(1000, size=64)
        self.MAX_LOADS_XYZAB = fixed(1064, size=5)
        self.gap04 = Gap(1069, size=11)
        self.RAW_ANALOG = fixed(1080, size=10)
        self.FILTERED_ANALOG = fixed(1090, size=8)
        self.SPINDLE_LOAD = one(1098)
        self.gap05 = Gap(1099, size=165)
        self.MAX_LOADS_CTUVW = fixed(1264, size=5)
        self.gap06 = Gap(1269, size=332)
        self.TOOL_TBL_FLUTES = tooltable(1601, size=200, typ="Int")
        self.TOOL_TBL_VIBRATION = tooltable(1801, size=200)
        self.TOOL_TBL_OFFSETS = tooltable(2001, size=200)
        self.TOOL_TBL_WEAR = tooltable(2201, size=200)
        self.TOOL_TBL_DROFFSET = tooltable(2401, size=200)
        self.TOOL_TBL_DRWEAR = tooltable(2601, size=200)
        self.gap07 = Gap(2801, size=199)
        self.ALARM = one(3000, typ="Int")
        self.T_MS = one(3001, typ="Time")
        self.T_HR = one(3002, typ="Time")
        self.SINGLE_BLOCK_OFF = ione(3003)
        self.FEED_HOLD_OFF = ione(3004)
        self.gap08 = Gap(3005, size=1)
        self.MESSAGE = ione(3006)
        self.gap09 = Gap(3007, size=4)
        self.YEAR_MONTH_DAY = one(3011, typ="Time")
        self.HOUR_MINUTE_SECOND = one(3012, typ="Time")
        self.gap10 = Gap(3013, size=7)
        self.POWER_ON_TIME = one(3020, typ="Time")
        self.CYCLE_START_TIME = one(3021, typ="Time")
        self.FEED_TIMER = one(3022, typ="Time")
        self.CUR_PART_TIMER = one(3023, typ="Time")
        self.LAST_COMPLETE_PART_TIMER = one(3024, typ="Time")
        self.LAST_PART_TIMER = one(3025, typ="Time")
        self.TOOL_IN_SPIDLE = ione(3026)
        self.SPINDLE_RPM = ione(3027)
        self.PALLET_LOADED = ione(3028)
        self.gap11 = Gap(3029, size=1)
        self.SINGLE_BLOCK = ione(3030)
        self.AGAP = one(3031)
        self.BLOCK_DELETE = ione(3032)
        self.OPT_STOP = ione(3033)
        self.gap12 = Gap(3034, size=162)
        self.TIMER_CELL_SAFE = one(3196, typ="Time")
        self.gap13 = Gap(3197, size=4)
        self.TOOL_TBL_DIAMETER = tooltable(3201, size=200)
        self.TOOL_TBL_COOLANT_POSITION = tooltable(3401, size=200)
        self.gap14 = Gap(3601, size=300)
        self.M30_COUNT1 = ione(3901)
        self.M30_COUNT2 = ione(3902)
        self.gap15 = Gap(3903, size=98)
        self.LAST_BLOCK_G = fixed(4001, size=21)
        self.gap16 = Gap(4022, size=79)
        self.LAST_BLOCK_ADDRESS = fixed(4101, size=26)
        self.gap17 = Gap(4127, size=874)
        self.LAST_TARGET_POS = cwcpos(5001)
        self.MACHINE_POS = machinepos(5021)
        self.WORK_POS = cwcpos(5041)
        self.SKIP_POS = cwcpos(5061)
        self.TOOL_OFFSET = fixed(5081, size=20)
        self.gap18 = Gap(5101, size=100)
        self.G52 = workoffsettable(5201)
        self.G54 = workoffsettable(5221)
        self.G55 = workoffsettable(5241)
        self.G56 = workoffsettable(5261)
        self.G57 = workoffsettable(5281)
        self.G58 = workoffsettable(5301)
        self.G59 = workoffsettable(5321)
        self.gap19 = Gap(5341, size=60)
        self.TOOL_TBL_FEED_TIMERS = tooltable(5401, size=100, typ="Secs")
        self.TOOL_TBL_TOTAL_TIMERS = tooltable(5501, size=100, typ="Secs")
        self.TOOL_TBL_LIFE_LIMITS = tooltable(5601, size=100, typ="Int")
        self.TOOL_TBL_LIFE_COUNTERS = tooltable(5701, size=100, typ="Int")
        self.TOOL_TBL_LIFE_MAX_LOADS = tooltable(5801, size=100)
        self.TOOL_TBL_LIFE_LOAD_LIMITS = tooltable(5901, size=100)
        self.gap20 = Gap(6001, size=197)
        self.NGC_CF = ione(6198)
        self.gap21 = Gap(6199, size=802)
        self.G154_P1 = workoffsettable(7001)
        self.G154_P2 = workoffsettable(7021)
        self.G154_P3 = workoffsettable(7041)
        self.G154_P4 = workoffsettable(7061)
        self.G154_P5 = workoffsettable(7081)
        self.G154_P6 = workoffsettable(7101)
        self.G154_P7 = workoffsettable(7121)
        self.G154_P8 = workoffsettable(7141)
        self.G154_P9 = workoffsettable(7161)
        self.G154_P10 = workoffsettable(7181)
        self.G154_P11 = workoffsettable(7201)
        self.G154_P12 = workoffsettable(7221)
        self.G154_P13 = workoffsettable(7241)
        self.G154_P14 = workoffsettable(7261)
        self.G154_P15 = workoffsettable(7281)
        self.G154_P16 = workoffsettable(7301)
        self.G154_P17 = workoffsettable(7321)
        self.G154_P18 = workoffsettable(7341)
        self.G154_P19 = workoffsettable(7361)
        self.G154_P20 = workoffsettable(7381)
        self.gap22 = Gap(7401, size=100)
        self.PALLET_PRIORITY = PalletTable(7501, size=100)
        self.PALLET_STATUS = PalletTable(7601, size=100)
        self.PALLET_PROGRAM = PalletTable(7701, size=100)
        self.PALLET_USAGE = PalletTable(7801, size=100)
        self.gap23 = Gap(7901, size=599)
        self.ATM_ID = ione(8500)
        self.ATM_PERCENT = one(8501, typ="Percent")
        self.ATM_TOTAL_AVL_USAGE = ione(8502)
        self.ATM_TOTAL_AVL_HOLE_COUNT = ione(8503)
        self.ATM_TOTAL_AVL_FEED_TIME = one(8504, typ="Secs")
        self.ATM_TOTAL_AVL_TOTAL_TIME = one(8505, typ="Secs")
        self.gap24 = Gap(8506, size=4)
        self.ATM_NEXT_TOOL_NUMBER = ione(8510)
        self.ATM_NEXT_TOOL_LIFE = one(8511, typ="Percent")
        self.ATM_NEXT_TOOL_AVL_USAGE = ione(8512)
        self.ATM_NEXT_TOOL_HOLE_COUNT = ione(8513)
        self.ATM_NEXT_TOOL_FEED_TIME = one(8514, typ="Secs")
        self.ATM_NEXT_TOOL_TOTAL_TIME = one(8515, typ="Secs")
        self.gap25 = Gap(8516, size=34)
        self.TOOL_ID = ione(8550)
        self.TOOL_FLUTES = ione(8551)
        self.TOOL_MAX_VIBRATION = one(8552)
        self.TOOL_LENGTH_OFFSETS = one(8553)
        self.TOOL_LENGTH_WEAR = one(8554)
        self.TOOL_DIAMETER_OFFSETS = one(8555)
        self.TOOL_DIAMETER_WEAR = one(8556)
        self.TOOL_ACTUAL_DIAMETER = one(8557)
        self.TOOL_COOLANT_POSITION = ione(8558)
        self.TOOL_FEED_TIMER = one(8559, typ="Secs")
        self.TOOL_TOTAL_TIMER = one(8560, typ="Secs")
        self.TOOL_LIFE_LIMIT = one(8561)
        self.TOOL_LIFE_COUNTER = one(8562)
        self.TOOL_LIFE_MAX_LOAD = one(8563)
        self.TOOL_LIFE_LOAD_LIMIT = one(8564)
        self.gap26 = Gap(8565, size=435)
        self.THERMAL_COMP_ACC = one(9000)
        self.gap27 = Gap(9001, size=15)
        self.THERMAL_SPINDLE_COMP_ACC = one(9016)
        self.gap28 = Gap(9017, size=983)
        self.GVARIABLES3 = fixed(10000, size=1000)
        self.INPUTS1 = fixed(11000, size=256)
        self.gap29 = Gap(11256, size=744)
        self.OUTPUT1 = fixed(12000, size=256)
        self.gap30 = Gap(12256, size=744)
        self.FILTERED_ANALOG1 = fixed(13000, size=13)
        self.COOLANT_LEVEL = one(13013)
        self.FILTERED_ANALOG2 = fixed(13014, size=50)
        self.gap31 = Gap(13064, size=936)
        self.SETTING = fixed(20000, size=10000)
        self.PARAMETER = fixed(30000, size=10000)

        self.TOOL_TYP = fixed(50001, size=200)
        self.TOOL_MATERIAL = fixed(50201, size=200)
        self.gap32 = Gap(50401, 50600)
        self.gap32 = Gap(51001, 51300)
        self.CURRENT_OFFSET = fixed(50601, size=200)
        self.CURRENT_OFFSET2 = fixed(50801, size=200)
        self.VPS_TEMPLATE_OFFSET = fixed(51301, size=100)
        self.WORK_MATERIAL = fixed(51401, size=200)
        self.VPS_FEEDRATE = fixed(51601, size=200)

        self.APPROX_LENGTH = fixed(51801, size=200)
        self.APPROX_DIAMETER = fixed(52001, size=200)
        self.EDGE_MEASURE_HEIGHT = fixed(52201, size=200)
        self.TOOL_TOLERANCE = fixed(52401, size=200)
        self.PROBE_TYPE = fixed(52601, size=200)

        self.PROBE = Alias(self.SKIP_POS)
        self.WORK = Alias(self.WORK_POS)

        self.MACHINE = Alias(self.MACHINE_POS)
        self.G53 = Alias(self.MACHINE_POS)


def txt_out(outname, names):
    guts = Table(
        title=f"{names.title} Macro Variables",
        caption=f"Generated by {__file__}",
    )

    guts.add_column("Range", justify="right")
    guts.add_column("N", justify="right")
    guts.add_column("K", justify="right")
    guts.add_column("Type", justify="center")
    guts.add_column("Name", justify="left")

    snames = sorted(names.interesting)
    for el in snames:
        guts.add_row(*el.for_txt())
        if el.alias:
            el = el.alias

    with lib.openw(outname) as out:
        console = rich.console.Console(file=out)
        console.print(guts, style=None)
        print("Generated ", outname)


def regen_out(outname, defs):
    with open(outname, "w", encoding="utf-8") as out:
        for el in sorted(defs.interesting):
            out.write("        " + el.for_regen())


def org_out(outname, defs):
    with lib.openw(outname) as out:
        out.write(
            "| <code>Name</code>          |  <code>Size</code> | <code>Address</code>         |\n"
        )
        out.write("| --- | --- | --- |\n")
        for el in sorted(defs.interesting):
            if el.name != "-":
                out.write(el.for_org())


def py_out(target_filename, defs):
    tmp_filepath = pathlib.Path(target_filename).with_suffix(".tmp")

    with open(target_filename, encoding="utf-8") as inf:
        repl = re.match(
            "(.*?# MACHINE GEN BELOW.*?).*(.*?# MACHINE.*)",
            inf.read(),
            flags=re.DOTALL,
        )
    if repl is not None:
        with open(tmp_filepath, "w", encoding="utf-8") as out:
            out.write(repl.group(1) + "\n")
            defs.py_write_to(out)
            out.write(INDENT + repl.group(2))
        tmp_filepath.rename(target_filename)


def makestdvars(outtxt_name, outdef_name, outpy_name, outorg_name):
    try:
        for names in [HaasNames()]:
            if outtxt_name:
                txt_out(outtxt_name, names)
            if outdef_name:
                regen_out(outdef_name, names)
            if outorg_name:
                org_out(outorg_name, names)
            if outpy_name:
                py_out(outpy_name, names)
    except FileNotFoundError as exc:  # no cover
        print(f"FAIL {exc.args[1]} '{exc.filename}'")
        return 1
    return 0



import enum

from p2g import *
from p2g.haas import *


# class ONumbers(enum.IntEnum):
#     PROBE_VICE_CENTER = 100
#     ALIGNC = enum.auto()
#     ROTARY_PROBE_BORE = enum.auto()
#     ABOVE_VICE = enum.auto()
#     ABOVE_ROTARY = enum.auto()
#     PROBE_CALIBRATE = enum.auto()


class Tool(enum.IntEnum):
    PROBE = 1
    KNOWN_LENGTH = 2


# pylint: disable=too-many-instance-attributes
class JobDefs:
    def __init__(self):
        super().__init__()
        self.MACHINE_ABS_ABOVE_OTS = Const(x=-1.16, y=-7.5, z=-7.5)

        # ots tool and probe are different lengths.
        self.MACHINE_ABS_PROBE_ABOVE_OTS = Const(x=-1.16, y=-7.5, z=-7.5)

        self.MACHINE_ABS_PROBE_ABOVE_RING = Const(x=-16.46, y=-3.5, z=-22.7)
        self.MACHINE_ABS_ABOVE_ROTARY = Const(x=-12.5214, y=-12.9896, z=-7.0)
        self.MACHINE_ABS_ABOVE_SEARCH_ROTARY_LHS_5X8 = Const(x=-15.5, y=-17.50, z=-14.0)
        self.MACHINE_ABS_ABOVE_VICE = Const(x=-28.0, y=-10.0, z=-16.00)
        self.MACHINE_ABS_CLOSE_ABOVE_OTS = Const(x=-1.16, y=-7.5, z=-7.6)
        self.MACHINE_ABS_CLOSE_ABOVE_TOOL_TOUCH = Const(
            x=-1.16,
            y=-7.5,
            z=-7.6,
        )
        self.MACHINE_ABS_HOME2 = Const(x=-15.0, y=-15.0, z=-2.0)
        self.MACHINE_ABS_ROTARY_HOME = Const(x=-12, y=0.0, z=-3.0)
        self.MACHINE_ABS_Z0 = Const(z=0.0)

        self.MACHINE_ABS_ZMIN = Const(z=-22.0)
        self.PROBE_RING_DIAMETER = 0.7
        m1 = Const(x=0.0, y=0.0, z=-0.7)
        self.MACHINE_ABS_SEARCH_ROTARY_LHS_5X8 = (
            self.MACHINE_ABS_ABOVE_SEARCH_ROTARY_LHS_5X8 + m1
        )

        self.FAST_FEED = 65.0
        self.SLOW_PROBE_FEED = 10.0
        self.FAST_PROBE_FEED = 10.0

        probe = goto.probe.mcode(MUST_SKIP).work
        self.slow_probe = probe.feed(self.SLOW_PROBE_FEED)

        self.fast_mabs_probe = probe.machine.absolute.feed(self.FAST_PROBE_FEED)
        self.slow_mabs_probe = probe.machine.absolute.feed(self.SLOW_PROBE_FEED)
        self.goto = goto.work.feed(self.FAST_FEED)

    def load_tool(self, tool):
        codenl(f"T{tool:02} M06")

    def setup_probing(self):
        self.load_tool(Tool.PROBE)

        codenl(PROBE_ON)
        codenl(NO_LOOKAHEAD)

    def pause(self, txt: str):
        comment("")
        message(MESSAGE, txt)
        comment("")

    def alarm(self, txt: str):
        comment("")
        message(ALARM, txt)
        comment("")

    def ots_on(self):
        codenl(OTS_ON)

    def spindle_probe_on(self):
        codenl(SPINDLE_PROBE_ON)

    def ots_calibrate(self):
        codenl("G65 P9023 A20. K5. S0.5 D-2.")

    def spindle_probe_find_height(self):
        codenl("G65 P9023 A21. T1.")

    def spindle_probe_find_radius(self):
        codenl("G65 P9023 A10.0 D0.7")



from typing import Optional

import p2g as pg


class SearchConstraint:
    # position roughly center above start of search.
    above: pg.Vec
    # min size of thing to look for.

    amin: pg.Vec
    # max size to search in one dimension for the thing
    # to probe.  pg.RVALnstains max size of things measured.
    amax: pg.Vec

    # when measuring z, how far to
    # move from an xy edge inwards to work it out.
    indent: pg.Vec

    # steps between probes

    delta: pg.Vec
    # safe distance above probed z surface to move
    skim_distance: pg.Vec

    # how far to backoff before probing slowly
    backoff: pg.Vec

    # how far to look down for a surface once we've
    # already located the rough z0
    search_depth: pg.Vec
    # how far to look down and not found is a not there
    #
    found_if_below: pg.Vec

    # to work out when there's a probe hit,
    # say probe from z down to 0,   then any stop < iota
    # is taken as a miss.
    iota: pg.Vec

    def __init__(
        self,
        amin: pg.Vec,
        amax: pg.Vec,
        above: pg.Vec,
        delta: pg.Vec,
        skim_distance=None,
        indent: Optional[pg.Vec] = None,
        backoff: Optional[pg.Vec] = None,
        search_depth=None,
        found_if_below=None,
        iota=None,
    ):
        super().__init__()

        if search_depth is None:
            search_depth = pg.Const(-0.1)

        if found_if_below is None:
            found_if_below = search_depth * 0.5

        if backoff is None:
            backoff = pg.Const(x=0.1, y=0.1, z=0.1)

        if indent is None:
            indent = round(amax * 0.1, 1)

        if skim_distance is None:
            skim_distance = pg.Const(0.3)

        if iota is None:
            iota = pg.Const(x=0.025, y=0.025, z=0.025)

        self.iota = iota
        self.delta = delta

        self.amin = amin
        self.indent = indent
        self.amax = amax

        self.above = above
        self.skim_distance = skim_distance
        self.backoff = backoff
        self.search_depth = search_depth
        self.found_if_below = found_if_below

    @property
    def comment(self):
        return [
            "Search Constraints:",
            "start:",
            f"  {self.above}",
            "boundary:",
            f"  x= (-{(self.amax[0] * 0.5)}..-{(self.amin[0] * 0.5)})"
            f"..({(self.amin[0] * 0.5)}..{(self.amax[0] * 0.5)})",
            f"  y= (-{(self.amax[1] * 0.5)}..-{(self.amin[1] * 0.5)})"
            f"..({(self.amin[1] * 0.5)}..{(self.amax[1] * 0.5)})",
            f"  z= ({(-self.amax[2])}..{(-self.amin[2])})",
            "indent:",
            f"  {self.indent}",
            "delta:",
            f"  {self.delta}",
        ]



import csearch
import defs

import p2g

from p2g.haas import *


# describe relationship between x&y and left,right,far,near
#


class Dinfo:
    name: str
    opposite: str
    cur_axis: int
    dxdy: tuple[float, float]

    def __init__(self, name, opposite, cur_axis, dxdy):
        self.name = name
        self.opposite = opposite
        self.cur_axis = cur_axis
        self.dxdy = dxdy


dinfo = {
    "left": Dinfo("left", "right", 0, (-1, 0)),
    "right": Dinfo("right", "left", 0, (1, 0)),
    "far": Dinfo("far", "near", 1, (0, 1)),
    "near": Dinfo("near", "far", 1, (0, -1)),
}


def prod_surface(st, sch, di, output):
    p2g.comment(
        "",
        f"quickly move probe to find {di.name} edge",
    )

    addr = p2g.base_addr()
    start_search = p2g.Const[2](sch.amin.xy * di.dxdy / 2.0)
    stop_search = p2g.Const[2](sch.amax.xy * di.dxdy / 2.0)
    delta = sch.delta * di.dxdy

    its = p2g.Var(
        (abs(stop_search - start_search) / sch.delta)[di.cur_axis] + 1,
    )
    cursor = st.cursor

    cursor[di.cur_axis] = start_search[di.cur_axis]
    st.goto.xy_then_z(cursor)
    while its > 0:
        st.goto.z_then_xy(cursor)
        st.fast_mabs_probe(z=sch.search_depth)
        if SKIP_POS.z < sch.search_depth + sch.iota:
            break

        cursor.xy += delta
        its -= 1
    else:
        st.alarm(f"search for {di.name} failed")

    p2g.comment(
        f"back off a bit to the {di.name}, then slowly probe ",
        f"{di.opposite}wards for precise measurement.",
    )
    cursor.xy += sch.backoff * di.dxdy
    st.goto(cursor.xy)
    st.slow_mabs_probe([0, 0])

    # sets part of tlc or brc
    output[di.cur_axis] = (SKIP_POS + PROBE_R * di.dxdy)[di.cur_axis]

    p2g.comment(
        "reposition above surface skim height,",
        f"just inside {di.name} edge",
    )

    st.goto.relative(sch.backoff.xy * di.dxdy)
    st.goto(z=sch.skim_distance)
    cursor.xy += -sch.indent.xy * di.dxdy
    st.goto(cursor)
    p2g.base_addr(addr)


def find_surface_before(st, sch):
    p2g.set_wcs(st.WCS)

    st.goto.machine(z=0)

    st.goto.machine.xy_then_z(sch.above)

    p2g.comment(f"find top z roughly set {st.WCS}.z.")

    st.WCS.xyz = MACHINE_POS.xyz
    # fast find move down to min search distance
    st.fast_mabs_probe(z=sch.amin.z)

    # make work offset z make rough top 0.
    st.WCS.z = MACHINE_POS.z
    st.pause("check g55")

    p2g.comment(
        "now work.z should be 0 at surface",
        "and work.xy roughly middle",
    )
    st.goto(z=sch.skim_distance)

    st.tlc = p2g.Var[2]()
    st.brc = p2g.Var[2]()

    st.cursor = p2g.Var[2](0, 0)
    prod_surface(st, sch, dinfo["left"], st.tlc)
    prod_surface(st, sch, dinfo["near"], st.brc)
    prod_surface(st, sch, dinfo["far"], st.tlc)

    # move back to centerline
    st.goto(0, 0)
    prod_surface(st, sch, dinfo["right"], st.brc)

    p2g.comment(
        " the 'error' between 0,0 and where we",
        " calculate the center to be gets",
        " added to cos and voila.",
    )

    st.error = p2g.Var[2]((st.tlc + st.brc) / 2.0)
    st.WCS.xy += st.error.xy
    st.goto(0, 0)

    p2g.comment(" final slow probe to find the surface z")
    st.slow_probe(z=sch.search_depth)
    st.WCS.z = SKIP_POS.z

    st.goto.machine(z=sch.above.z)
    st.alarm(" what changed")


def runit(st, sch):
    st.WCS = G55

    p2g.comment(
        "Find center of plate in vice,",
        f" result in {st.WCS}",
        *sch.comment,
    )

    st.setup_probing()

    find_surface_before(st, sch)


def vicecenter():
    st = defs.JobDefs()
    p2g.symbol.Table.print = 1
    sch = csearch.SearchConstraint(
        # minimum size expected
        amin=p2g.Const(x=7.0, y=4.0, z=-5.0),
        # maximum size expected
        amax=p2g.Const(x=14.0, y=8.0, z=3.0),
        delta=p2g.Const(x=0.75, y=0.4),
        above=st.MACHINE_ABS_ABOVE_VICE + p2g.Const(x=0.0, z=0.0, y=-3.0),
    )
    runit(st, sch)



import contextlib
import pathlib
import sys


# open accouting for '-' files.
# sourcing empty file when in test.


@contextlib.contextmanager
def openw(name):
    if str(name) == "-":
        yield sys.stdout
    else:
        with open(name, "w", encoding="utf-8") as wfile:
            yield wfile


def write_nl_lines(thing, outfile_path):
    with openw(outfile_path) as outf:
        for line in thing:
            print(line, file=outf)


def same(lhs, rhs):
    if type(lhs) is not type(rhs):
        return False

    return lhs.same(rhs)


def nljoin(line_list):
    return "\n".join(line_list)


# unwind a list of lists etc can turn generators of lists of strings
# into space joined string.


def unwind1(args):
    if isinstance(args, str):
        yield args
    else:
        for el in args:
            yield from unwind1(el)


def unwind(args):
    return " ".join(unwind1(args))


# find the file if in dist or just checkout out.
def find_ours(filename):
    for dir in [".", ".."]:
        for subdir in ["tests", "doc", "examples"]:
            look_here = pathlib.Path(__file__).parent / dir / subdir / filename

            if look_here.exists():
                return look_here.resolve()

    raise FileNotFoundError  # no cover



import typing

from p2g import axis
from p2g import err
from p2g import nd
from p2g import scalar
from p2g import stat


class Vec(nd.EBase):
    is_scalar = False
    is_vec = True

    _assoc_name: str
    _read: bool

    def __init__(self):
        self._read = False
        self._assoc_name = ""

    def to_scalar(self):
        return self[0]

    def get_axes_vec(self, key):
        if key == "var":
            return RValueVec(list(scalar.urange(0, self.nelements())))

        indexes = axis.name_to_indexes_list(key)
        if max(indexes) >= self.nelements():
            err.compiler("Reference to too many axes.")
        return RValueVec(indexes)

    def forever(self):
        self._read = True
        while True:
            yield from self.everything()

    def __getattr__(self, key):
        if key[0] == "_":
            return getattr(self, key)

        axes_vec = self.get_axes_vec(key)

        return sorv_from_list([self.get_at(idx) for idx in axes_vec])

    def __setattr__(self, key, value):
        if key[0] == "_":
            object.__setattr__(self, key, value)
            return
        value = wrap_maybe_vec(value)
        axes_vec = self.get_axes_vec(key)

        for idx, src in zip(axes_vec, value.forever()):
            self.set_at(idx, src)

    def __getitem__(self, scalar_index):
        if isinstance(scalar_index, slice):
            return self.get_slice(scalar_index)
        return self.get_at(scalar.wrap_scalar(scalar_index))

    def __add__(self, _other) -> "Vec":
        return self  # no cover, filled in from op.reg

    def __mul__(self, _other) -> "Vec":
        return self  # no cover, filled in from op.reg

    def __lt__(self, _other) -> "Vec":
        return self  # no cover, filled in from op.reg


class RValueVec(Vec):
    _guts: list[scalar.Scalar]
    _from_user: bool

    # def __eq_(self, a):
    #     breakpoint()

    def __init__(self, value, from_user=False):
        super().__init__()

        object.__setattr__(self, "_guts", [])

        object.__setattr__(self, "_from_user", from_user)
        for el in value:
            self._guts.append(scalar.wrap_scalar(el))

    @property
    def user_defined(self):
        return self._from_user

    def get_address(self):
        return self._guts[0].get_address()

    def everything(self):
        self._read = True
        return self._guts

    def get_at(self, idx: scalar.Scalar):
        self._read = True
        return self._guts[int(idx)]

    def nelements(self):
        return len(self._guts)

    def to_symtab_entry(self, _) -> str:
        res = []
        tlen = 0
        for el in self._guts:
            nextstr = nd.to_gcode(el, nd.NodeModifier.F3X3)
            tlen += len(nextstr)
            if tlen > 21:
                res.append("...")
                break
            res.append(nextstr)

        return ",".join(res)

    def __iter__(self):
        yield from self._guts

    def __str__(self):
        return ", ".join(map(str, self._guts))


class MemVec(Vec):
    # nasty fix forward ref
    make_hashop: typing.Callable

    _size: scalar.Scalar
    _addr: scalar.Scalar
    _step: scalar.Scalar

    # $    def __eq__(self, x):
    #        breakpoint()
    #    def __hash__(self):
    #        return 1

    def __init__(self, _addr=0, _size=0, _step=1):
        super().__init__()
        self._addr = scalar.wrap_scalar(_addr)
        self._size = scalar.wrap_scalar(_size)
        self._step = scalar.wrap_scalar(_step)

    @property
    def user_defined(self):
        return True

    def get_address(self):
        return self._addr

    def to_symtab_entry(self, varrefs) -> str:
        fwidth = 7
        hit_indexes = {}

        for el, addr in enumerate(scalar.urange(self._addr, self._addr + self._size)):
            if addr in varrefs:
                hit_indexes[el] = True

        if hit_indexes and max(hit_indexes) >= len(axis.NAMES):
            return f"#{self._addr}[{self._size}]".rjust(fwidth)
        res = []

        for idx in scalar.urange(0, self._size):
            if idx not in hit_indexes:
                continue

            res.append(
                f"#{self._addr + idx}.{axis.low_names(idx)}".rjust(fwidth),
            )
        return " ".join(res)

    def nelements(self):
        return int(self._size)

    def everything(self):
        return (
            self.get_at(scalar.wrap_scalar(idx))
            for idx in scalar.urange(0, self._size, int(self._step))
        )

    def get_at(self, idx: scalar.Scalar):
        if isinstance(idx, scalar.Constant):
            fidx = int(idx)
            if not 0 <= fidx < int(self._size):
                err.compiler(
                    f"Index out of range, index={fidx} size={self._size}.",
                )

        return MemVec.make_hashop(idx, self._addr)

    def set_at(self, idx: scalar.Scalar, src: scalar.Scalar):
        stat.append_set(self.get_at(idx), src)

    def get_slice(self, index: slice):
        tmp = (list(range(self.nelements())))[index]
        step = 0
        if len(tmp) > 1:
            step = tmp[1] - tmp[0]

        addr = self._addr + tmp[0]
        size = tmp[-1] - tmp[0] + 1
        assert addr is not None
        return MemVec(addr, size, step)

    def __setitem__(self, indexes, src):
        if isinstance(indexes, slice):
            indexes = list(range(*indexes.indices(self.nelements())))
        indexes = wrap_maybe_vec(indexes)
        src = wrap_maybe_vec(src)
        for idx, sel in zip(indexes.everything(), src.forever()):
            stat.append_set(self[idx], sel)

    def __repr__(self):
        return f"(array  {self._addr} {self._size})"


def wrap_optional_maybe_vec(
    thing,
) -> typing.Optional[typing.Union[Vec, scalar.Scalar]]:
    if (res := scalar.wrap_optional_scalar(thing)) is not None:
        return res

    if isinstance(thing, list):
        return RValueVec(thing)
    if isinstance(thing, (RValueVec, MemVec)):
        return thing
    if isinstance(thing, (set, dict)):
        return RValueVec(thing)
    if isinstance(thing, tuple):
        return RValueVec(list(thing))

    raise NotImplementedError


def wrap_maybe_vec(thing) -> typing.Union[Vec, scalar.Scalar]:
    res = wrap_optional_maybe_vec(thing)
    assert res is not None
    return res


def sorv_from_list(thing: list) -> typing.Union[Vec, scalar.Scalar]:
    if len(thing) != 1:
        return RValueVec(thing)
    return thing[0]



#! /usr/bin/env python
import datetime
import pathlib
import re
import shutil
import sys

import docopt

import p2g

from p2g import err
from p2g import gbl
from p2g import lib
from p2g import makestdvars
from p2g import walk


DOC = """
Turns a python program into a gcode program.

Usage:
   p2g [--function=<fname> ]
       [--job=<jobname>]
       [options] gen <srcfile> [<outfile>]
#
#       Read from python <srcfile>, emit G-Code.
#
#         Output file name may include {time} which will create a decrementing
#         prefix for the output file which makes looking for the .nc in a
#         crowded directory simpler.
#
#       Examples:
#          p2g gen foo.py ~/_nc_/{time}O001-foo.nc
#             Makes an output of the form ~/_nc_/001234O001-foo.nc
#
#          p2g gen --func=thisone -
#             Read from stdin, look for the 'thisone' function and write to
#             to stdout.
#
   p2g examples <exampledir>
#         Create <exampledir>, populate with examples and compile.
#
#         Examples:
#           p2g examples showme
#             Copies the examples into ./showme and then runs
#              p2g gen showme/vicecenter.py
#              p2g gen showme/checkprobe.py
#
   p2g doc
#         Send readme.txt to console.
#
   p2g help
#         Show complete command line help.
#
   p2g version
#         Show version.
#
   p2g location
#         Show which p2g is running.
#
#      For maintenance:
     p2g [options] stdvars [--txt=<txt>] [--dev=<dev>]
                           [--py=<py>] [--org=<org>]
#               Recreate internal files.

 Options:
  --narrow                    Narrow output; formatted to fit in the
                              narrow space of the CNC machine's program
                              display.
#
#      For maintenance:
           --no-boiler-plate  Turn of job entry and terminal M30.
           --break            Breakpoint on error.
           --debug            Enter debugging code.
           --verbose          Too much.
           --logio            Even more.


"""


def do_examples(outdir):
    find_examples = lib.find_ours("vicecenter.py")
    dst_dir = outdir.resolve()
    example_dir = find_examples.parent
    example_files = example_dir.glob("[a-z]*.py")

    outdir.mkdir(exist_ok=True, parents=True)

    for src in example_files:
        dst_name = outdir / src.name
        gbl.sprint(f"Copying {src} {dst_name}")
        shutil.copy(src, dst_name)

    for job in ["vicecenter", "probecalibrate"]:
        rootname = dst_dir / job
        sysargs = [
            "gen",
            rootname.with_suffix(".py"),
            rootname.with_suffix(".nc"),
        ]

        gbl.sprint(f"Running {sysargs[0]} {sysargs[1]} {sysargs[2]}")
        recur(sysargs)


def do_doc():
    doc = lib.find_ours("readme.txt").read_text()
    gbl.sprint(doc.replace("</code>", "]").replace("<code>", "["))


def calculate_output_file_name(provided_filename):
    # if provided_filename is empty then output is stdout.
    if provided_filename is None:
        return "-"

    now = datetime.datetime.now()
    prev_midnight = now.replace(hour=0, minute=0, second=0, microsecond=0)
    next_midnight = 24 * 60
    mins_togo = next_midnight - (now - prev_midnight).seconds // 60

    return provided_filename.replace("{time}", f"{mins_togo:05d}")


def do_gen(src_name, job_name, func_name, output_name):
    if func_name is None:
        func_name = "<last>"

    job_name = job_name if job_name else "O0001"

    output_name = calculate_output_file_name(output_name)
    src_path = pathlib.Path(src_name)

    gbl.log(f"src: {src_path}")
    gbl.log(f"fnc: {func_name}")
    gbl.log(f"job: {job_name}")
    gbl.log(f"output: {output_name}")
    try:
        res = walk.compile2g(func_name, src_path, job_name=job_name)
        lib.write_nl_lines(res, output_name)
        return 0
    except err.CompilerError as exn:
        exn.report_error(absolute_lines=True)
    return 1


def inner_main(options: list[str]):
    # remove comments from source to docopt.

    parseable_opts = re.sub("#.*\n", "", DOC)
    options = [str(v) for v in options]

    opts = docopt.docopt(parseable_opts, help=False, argv=options)

    gbl.config = gbl.config._replace(
        debug=opts["--debug"],
        bp_on_error=opts["--break"],
        verbose=(opts["--verbose"]),
        logio=(opts["--logio"]),
    )

    res = 0
    # find first options which are set and start with a letter - ie
    # the command - there can only be one.
    match [k for (k, v) in opts.items() if v and k[0].isalpha()]:
        case [] | ["help"]:
            gbl.sprint(DOC.strip("\n").replace("#", " "))

        case ["version"]:
            gbl.sprint(f"Version: p2g {p2g.VERSION}")

        case ["location"]:
            gbl.sprint(f"{pathlib.Path(__file__).parent}")

        case ["examples"]:
            do_examples(pathlib.Path(opts["<exampledir>"]))

        case ["stdvars"]:
            makestdvars.makestdvars(
                opts["--txt"],
                opts["--dev"],
                opts["--py"],
                opts["--org"],
            )

        case ["doc"]:
            do_doc()

        case ["gen"]:
            gbl.config = gbl.config._replace(
                narrow_output=gbl.config.narrow_output or opts["--narrow"],
                boiler_plate=gbl.config.boiler_plate and not opts["--no-boiler-plate"],
            )

            res = do_gen(
                src_name=opts["<srcfile>"],
                job_name=opts["--job"],
                func_name=opts["--function"],
                output_name=opts["<outfile>"],
            )

    return res


def main():
    with gbl.save_config(narrow_output=False):
        return inner_main(sys.argv[1:])


def recur(options: list[str]):
    with gbl.save_config(narrow_output=False, tin_test=True, boiler_plate=True):
        return inner_main(options)


def nrecur(options: list[str]):
    with gbl.save_config(narrow_output=False, tin_test=False, boiler_plate=False):
        return inner_main(options)



# magic for things like
#  a) g.goto.fast ([1,2])
#  b) foo = g.probe.slow
#  c)foo(z=12)


import dataclasses
import typing

from p2g import axis
from p2g import err
from p2g import gbl
from p2g import scalar
from p2g import stat
from p2g import vector


def flatten(args):
    res = []
    for el in args:
        for subel in el.everything():
            res.append(subel)
    return res


def unpack(args, kwargs) -> vector.RValueVec:
    resmap: typing.Dict[int, scalar.Scalar] = {}
    if args:
        args = map(vector.wrap_maybe_vec, args)
        args = flatten(args)

        for idx, value in enumerate(args):
            resmap[idx] = value

    # can have x=1,y=1 or xy=(somthing)
    for axis_string, values in kwargs.items():
        values = vector.wrap_maybe_vec(values)
        axis_indexes = axis.name_to_indexes_list(axis_string)
        for axis_idx, value in zip(axis_indexes, values.forever()):
            if axis_idx in resmap:
                err.compiler(f"Overlapping axes {args} {kwargs}.")
            resmap[axis_idx] = value

    # fill any holes.
    coord_size = 0
    if resmap:
        coord_size = max(resmap) + 1

    # rules:
    # everything in the input gets packed into
    # the output in the right order.
    # if something is a nan in the input, and
    # we've got gap_default, then the nan is
    # turned into that.

    min_size = 0
    return vector.RValueVec(
        [resmap.get(x) for x in range(max(min_size, coord_size))], from_user=True
    )


# build things like AABuilder(1,2,3, _addr=123) and AABuilder[8](_addr=123)
@dataclasses.dataclass
class _TypeBuilder:
    btype: str
    size: typing.Optional[int]

    def __init__(self, btype="c", *, size=None):
        self.btype = btype
        self.size = size

    def check_addr_sanity(self, addr):
        if addr is None:
            return
        if self.btype == "c":
            err.compiler("Const can't have an address.")
        elif self.btype == "a":
            err.compiler("Var can't have an address.")

    def check_size_sanity(self, size, initialized_size):
        if initialized_size > 0 and initialized_size != size:
            err.compiler(f"Conflicting sizes {self.size} and {initialized_size}.")
        if size == 0 and initialized_size == 0:
            err.compiler("Zero sized vector.")

    def workout_size(self, initialized_size):
        size = self.size
        if size is None:
            size = max(initialized_size, 1)
        self.check_size_sanity(size, initialized_size)
        return size

    def __call__(
        self,
        *args,
        addr=None,
        **kwargs,
    ):
        self.check_addr_sanity(addr)
        values = unpack(args, kwargs)

        initialized_size = values.nelements()

        size = self.workout_size(initialized_size)

        if self.btype == "c":
            return values
        if addr is None:
            addr = gbl.iface.next_bss(size)
        res = vector.MemVec(addr, size)
        if values.nelements():
            for lhs, rhs in zip(res.everything(), values.forever()):
                stat.append_set(lhs, rhs)

        return res

    # sets vector size.
    def __getitem__(self, el):
        return _TypeBuilder(
            size=el,
            btype=self.btype,
        )


Var = _TypeBuilder("a")
Fixed = _TypeBuilder("f")
Const = _TypeBuilder("c")



import functools
import inspect
import itertools
import pathlib
import typing

from p2g import err
from p2g import gbl
from p2g import lib
from p2g import walk


# takes fn, works out where it lives, where to find golden output and
# where to put test output
# test_foo.py (test_zap) will make a tests/golden/test_foo_test_zap.nc file.


def make_file_path(func, new_suffix) -> pathlib.Path:
    # this_file_path <- <somewhere>/ptest.py
    this_file_path = pathlib.Path(__file__)
    # this_file_directory <- <somewhere>/p2g
    this_file_directory = this_file_path.parent

    # file_part <- function's filename
    file_part = str(pathlib.Path(func.__code__.co_filename).stem)
    #    file_part = file_part.replace("test_", "")

    # can be in <srcdir>/tests or ../tests depending upon
    # if run installed or not. Look for testdir.

    for tests_dir in [
        this_file_directory / "tests",
        this_file_directory.parent / "tests",
    ]:
        if (tests_dir / "golden").exists():
            break
    else:  # pragma: no cover
        breakpoint()

        err.compiler("can't find tests")

    # generated_dir <- <somewhere>/tests/golden
    generated_dir = tests_dir / "golden"
    generated_dir.mkdir(exist_ok=True)
    #  <somewhere>/tests/golden/foo_zap.nc

    #    new_filename = func.__name__.replace("test", file_part)
    new_filename = file_part + "_" + func.__name__
    return (generated_dir / new_filename).with_suffix(new_suffix)


def strip_comments(txt):
    txt = txt.replace(" ", "")
    idx = txt.find("(")
    if idx >= 0:
        return txt[:idx].strip()
    return txt.strip()


# find differences between two str lists
# line where error found, and a list of *s assocaited with
# broken lines.

SAME_MARKER = " | "
DIFF_MARKER = " * "


def strip_lines_comments(block):
    for line in block:
        yield strip_comments(line)


# find differences in stuff which isn't commented.
# return a set of lines with diffs.
def find_differences(
    golden_data: typing.Sequence[str],
    callow_data: typing.Sequence[str],
    check_comments: bool,
):
    lhs_src = golden_data
    rhs_src = callow_data
    if not check_comments:
        lhs_src = strip_lines_comments(lhs_src)
        rhs_src = strip_lines_comments(rhs_src)

    diffs = set()
    # skip same stuff
    for line, (want, got) in enumerate(
        itertools.zip_longest(lhs_src, rhs_src, fillvalue="")
    ):
        if want != got:
            diffs.add(line)
    return diffs


@gbl.g2l
def format_differences(marks, golden_data: list[str], callow_data: list[str]):
    fromto = slice(max(min(marks) - 4, 0), min(marks) + 10)

    lhs_width = max(len(x) for x in golden_data[fromto])

    def fmt(lhs, middle, rhs):
        return lhs.ljust(lhs_width) + middle + rhs

    yield fmt("want", "", "got")

    for idx, (lhs, rhs) in enumerate(
        itertools.zip_longest(
            golden_data[fromto],
            callow_data[fromto],
            fillvalue=" - ",
        )
    ):
        yield fmt(
            lhs,
            DIFF_MARKER if idx + fromto.start in marks else SAME_MARKER,
            rhs,
        )


def writelines(fn, suffix, txt, comment):
    path = make_file_path(fn, suffix)
    print(f"WRRTIGIN {comment} to ", path)
    gbl.log(f"Ptest error, writing to {comment}")
    return path.write_text(lib.nljoin(txt))


# when a test fails, create source which would
# have passed.


def make_decorated_source_seed(fn, callow_data):
    tofix = ["@must_be("]
    for line in callow_data:
        quotechar = '"'
        if quotechar in line:
            quotechar = "'"
        if line:
            tofix.append(quotechar + line + quotechar + ",")
    tofix.append(")\n")

    lines = gbl.splitnl(inspect.getsource(fn))

    while lines and not lines[0].startswith("def"):
        lines = lines[1:]

    writelines(fn, ".decorator", tofix + lines, "creating decorator")


# compile fn, return generated errors or text.
@gbl.g2l
def get_all_comp_outputs(fn):
    try:
        outlines = walk.compile2g(fn.__name__, inspect.getsourcefile(fn), job_name=None)
    except err.CompilerError as exn:
        outlines = exn.error_lines(absolute_lines=False, topfn=fn)

    return outlines


@gbl.g2l
def read_and_trim(path):
    lines = gbl.splitnl(path.read_text())
    for line in lines:
        yield line


# when called from test, exceptions are sent to output file
# as well as being passed up.


# compile code and compare output with file.


def check_must_be_worker(fn, gold_data, check_comments=False):
    callow_data = list(get_all_comp_outputs(fn))

    markers = find_differences(gold_data, callow_data, check_comments)

    if not markers:
        return
    etext = format_differences(markers, gold_data, callow_data)

    make_decorated_source_seed(fn, callow_data)

    if "FORCE" in gold_data[0]:
        return

    # if running in pytest then exist out gracefully for them.
    # otherwise, running in debug harness.
    if gbl.config.debug:  # for debug
        err.compiler(f"ptest error {lib.nljoin(etext)}")

    assert False, lib.nljoin(etext)  # for debug


def check_golden_worker(fn, check_comments):
    callow = list(get_all_comp_outputs(fn))

    gold_path = make_file_path(fn, ".nc")
    if gold_path.exists():
        gold_data = list(read_and_trim(gold_path))

        markers = find_differences(gold_data, callow, check_comments)
        if not markers:
            return
        etext = format_differences(markers, gold_data, callow)

        writelines(fn, ".got", callow, "compare fail so")

        # if running in pytest then exist out gracefully for them.
        # otherwise, running in debug harness.
        if gbl.config.debug:  # for debug
            for line in etext:
                print(line)
            err.compiler(f"ptest error {lib.nljoin(etext)}")

        assert False, lib.nljoin(etext)

    else:
        # no source gold, make it.
        writelines(fn, ".nc", callow, "no source gold, create")


# decorator for tests, turns the node into ast and calls
# the output comparer, uses relative linenumbers in reports
# so things can be added without breaking exisiting


def check_golden(check_comments=False):
    def check_golden_(fn):
        @functools.wraps(fn)
        def check_golden__():
            check_golden_worker(fn, check_comments)

        return check_golden__

    return check_golden_


def check_golden_nodec(fn):
    @functools.wraps(fn)
    def check_golden__():
        check_golden_worker(fn, False)

    return check_golden__


def must_be(*text):
    def must_be_(fn):
        @functools.wraps(fn)
        def must_be__():
            check_must_be_worker(fn, text)

        return must_be__

    return must_be_


def must_be_cc(*text):
    def must_be_(fn):
        @functools.wraps(fn)
        def must_be__():
            check_must_be_worker(fn, text, check_comments=True)

        return must_be__

    return must_be_


def init_for_pytest():
    gbl.config = gbl.config._replace(
        narrow_output=True,
    )



#! /usr/bin/env python
import sys

import defs

from p2g import *


def probecalibrate():
    st = defs.JobDefs()
    if 1:
        comment(
            "Start with fixed height probe,",
            "make sure probe stickout <2.25in",
        )

        Table.print = 1
        st.load_tool(defs.Tool.KNOWN_LENGTH)
        st.ots_on()
        st.goto.machine(z=0)
        st.pause("touch OTS, must beep")
        st.goto.machine.xy_then_z(st.MACHINE_ABS_ABOVE_OTS)

        st.pause("Make sure tool position looks safe.")
        st.goto.machine(z=st.MACHINE_ABS_CLOSE_ABOVE_OTS.z)
        st.ots_calibrate()
    if 1:
        comment("Calibrate spindle probe.")

        st.load_tool(defs.Tool.PROBE)
        st.spindle_probe_on()

        st.pause("touch probe, must beep")

        comment("test spindle probe with OTS.")
        st.goto.machine(z=0)
        st.goto.machine.xy_then_z(st.MACHINE_ABS_PROBE_ABOVE_OTS)

        st.spindle_probe_find_height()

        comment("test spindle probe with ring.")
        st.goto.machine(z=0)
        st.goto.machine.xy_then_z(st.MACHINE_ABS_PROBE_ABOVE_RING)
        st.spindle_probe_find_radius()
        st.goto.machine(z=0)



import abc
import dataclasses
import itertools
import typing

from p2g import err
from p2g import gbl
from p2g import lib
from p2g import nd


COMMENT_INDENT = 34


# infront of code which isn't a label
NORMAL_PREFIX = "  "


@dataclasses.dataclass
class Label:
    idx: int
    used: bool

    def __init__(self, idx: int):
        self.idx = idx
        self.used = False

    def as_gcode_ref(self):
        return f"{self.idx}"

    def as_gcode_definition(self):
        return f"N{self.idx}"


# auto comment is a comment which comes from
# the source, rather than being typed.
# we take liberties to increase information density.


# remove bad chars from a comment.
def clean_comment_chars(txt: str):
    res = ""
    for ch in txt:
        match ch:
            case "(":
                ch = "["
            case ")":
                ch = "]"

        res += ch
    return res


def compress_and_clean(line: str):
    guts = clean_comment_chars(line)
    for too_talky in ["p2g.", "var."]:
        guts = guts.replace(too_talky, "")

    if guts.startswith("    "):
        guts = guts[4:]

    return "( " + guts.ljust(30) + ")"


@dataclasses.dataclass
class StatBase(abc.ABC):
    _comment: str

    def __init__(self, comment_txt=None):
        self.pos = err.state.last_pos

        # if specifically asked for none, then we're done.
        if comment_txt == "<none>":
            self._comment = ""
        else:
            # no comment supplied, then use source line.
            if not comment_txt:
                comment_txt = err.src_code_from_node_place(self.pos)
            comment_txt = compress_and_clean(comment_txt)
            self._comment = comment_txt

    def to_line_lhs(self) -> list[str]:
        raise AssertionError

    @gbl.g2l
    def to_full_lines(self, blockstate):
        comtxt = self._comment
        if comtxt == blockstate.prev_comtxt:
            comtxt = ""
        else:
            blockstate.prev_comtxt = comtxt

        for code_txt, com_txt in itertools.zip_longest(
            self.to_line_lhs(),
            [comtxt],
            fillvalue="",
        ):
            # If code would leak into comment, put out the comment on
            # own line.  If in narrow mode (by option,  or default if
            # in pytest or debug), put on own line too.

            if gbl.config.narrow_output:
                comment_indent = 0
            else:
                comment_indent = COMMENT_INDENT

            code_comment_gap = comment_indent - len(code_txt)
            if code_comment_gap < 0:
                first_line = com_txt
                second_line = code_txt
            else:
                if com_txt:
                    first_line = code_txt + " " * code_comment_gap + com_txt
                else:
                    first_line = code_txt
                second_line = ""
            if first_line:
                yield first_line
            if second_line:
                yield second_line


@dataclasses.dataclass
class Nest:
    first_label: int
    next_label: int

    prev: "Nest"
    slist: list[StatBase]
    cur: "Nest" = typing.cast("Nest", None)

    def __init__(self):
        self.prev = Nest.cur
        if gbl.config.tin_test or not Nest.cur:
            self.first_label = 1000
        else:
            self.first_label = self.prev.first_label + 1000

        Nest.cur = self
        self.next_label = self.first_label
        self.slist = []

    def get_label(self):
        res = self.next_label
        self.next_label += 1
        return Label(res)

    @classmethod
    def add_stat(cls, stat):
        Nest.cur.slist.append(stat)

    @gbl.g2l
    def to_full_lines(self):
        class BS:
            prev_comtxt = ""

        blockstate = BS()
        for line in self.slist:
            yield from line.to_full_lines(blockstate)

    def __enter__(self):
        return self

    def __exit__(self, *_exn):
        Nest.cur = self.prev


class CommentLines(StatBase):
    def __init__(self, lines):
        super().__init__()
        self.lines = lines

    def to_full_lines(self, _):
        lines = list(map(str, self.lines))

        def max_str_len(lines):
            return len(max(lines, key=len, default=""))

        def pad_to_same_width(lines):
            wid = max_str_len(lines)
            res = []
            for line in lines:
                res.append(line.ljust(wid))
            return res

        lines = pad_to_same_width(lines)

        for line in lines:
            if line.strip():
                yield "( " + clean_comment_chars(line) + " )"
            else:
                yield ""


def add_stat(stat):
    gbl.log(f"Adding {stat}")
    Nest.add_stat(stat)


def comment(*lines):
    add_stat(CommentLines(" "))
    add_stat(CommentLines(lines))


def com(*lines):
    add_stat(CommentLines(lines))


@dataclasses.dataclass
class Goto(StatBase):
    target: Label

    def __init__(self, target: Label):
        super().__init__()
        self.target = target

    def to_line_lhs(self):
        yield f"  GOTO {self.target.as_gcode_ref()}"


@dataclasses.dataclass
class If(StatBase):
    exp: typing.Any
    on_t: Label

    def __init__(self, exp, on_t: Label):
        super().__init__()

        self.exp = exp
        self.on_t = on_t

    def to_line_lhs(self):
        lhs = f"IF [{nd.to_gcode(self.exp)}] "
        rhs = f"GOTO {self.on_t.as_gcode_ref()}"
        yield NORMAL_PREFIX + lhs + rhs


@dataclasses.dataclass
class Code(StatBase):
    txt: str

    def __init__(self, txtargs, comment_txt=""):
        super().__init__(comment_txt=comment_txt)

        self.txt = lib.unwind(txtargs)

    def to_line_lhs(self):
        yield NORMAL_PREFIX + self.txt


@dataclasses.dataclass
class Dprint(StatBase):
    txt: str

    def __init__(self, txt):
        super().__init__("<none>")
        self.txt = txt

    def to_line_lhs(self):
        yield "DPRNT[" + self.txt + "]"


class Set(StatBase):
    lhs: nd.EBase
    rhs: nd.EBase

    def __init__(self, lhs: nd.EBase, rhs: nd.EBase, comtxt: str = ""):
        super().__init__(comtxt)
        assert isinstance(lhs, nd.EBase)
        self.lhs = lhs
        self.rhs = rhs

    @gbl.g2l
    def to_line_lhs(self):
        lhs = self.lhs.to_gcode(nd.NodeModifier.EMPTY)
        rhs = self.rhs.to_gcode(nd.NodeModifier.ARGUMENT)
        yield f"{NORMAL_PREFIX}{lhs}= {rhs}"


def append_set(dst, src, comtxt=""):
    if not lib.same(dst, src):
        add_stat(Set(dst, src, comtxt))


def code(*txtargs, comment_txt: str = ""):
    add_stat(Code(txtargs, comment_txt=comment_txt))


def codenl(txtlst, comtxt: str = ""):
    if isinstance(txtlst, str):
        add_stat(Code(txtlst, comtxt))
    else:
        for txt in txtlst:
            codenl(txt, comtxt)
            comtxt = ""


# take list of args and chain them
# and then join. Take care with strings.


@dataclasses.dataclass
class LabelDef(StatBase):
    labeldef: Label

    def __init__(self, labeldef: Label):
        self.labeldef = labeldef
        super().__init__()

    def to_line_lhs(self):
        yield f"{self.labeldef.as_gcode_definition()}"


def next_label():
    return Nest.cur.get_label()


def dprint(fstr):
    fstr = fstr.replace(" ", "*")
    add_stat(Dprint(fstr))


# when run from tests, sometimes we get generated
# statements by surprise.
top = Nest()



import abc
import contextlib
import functools
import sys
import typing


class Config(typing.NamedTuple):
    debug: bool = False
    narrow_output: bool = False
    bp_on_error: bool = False
    boiler_plate: bool = True
    verbose: bool = False
    logio: bool = False
    # overwrittern if we get in via main
    tin_test: bool = True


config = Config()


@contextlib.contextmanager
def save_config(**kwargs):
    global config  # pylint: disable=global-statement
    old = config
    config = config._replace(**kwargs)
    yield

    config = old


class PerTranslation:
    ebss: int
    varrefs: typing.Dict
    last_node: typing.Any

    def __init__(self):
        self.reset()

    def next_bss(self, size):
        addr = self.ebss
        self.ebss += size
        return addr

    def reset(self):
        self.varrefs = {}
        self.ebss = 100

        self.last_node = None


iface = PerTranslation()


def log(*args):
    if config.verbose:
        print(*args)


def sprint(*args):
    print(*args)


# print to redirected stderr
def eprint(*args):
    print(*args, file=sys.stderr)


@functools.cache
def logread(handle):
    res = handle.read()
    if config.logio:
        print(res)  # no cover
    return res


def splitnl(line):
    return line.split("\n")


######################################################################
# i/o redirection


@contextlib.contextmanager
def openr(name):
    if str(name) == "-":
        yield sys.stdin, None
    else:
        try:
            with open(name, "r", encoding="utf-8") as rfile:
                yield rfile, None
        except FileNotFoundError as err:
            yield None, err


def g2l(generator):
    def g2l_(*args, **kwargs):
        return list(generator(*args, **kwargs))

    return g2l_


class HasToSymTab(abc.ABC):
    user_defined = False

    @abc.abstractmethod
    def to_symtab_entry(self, _addrs_used) -> str:
        return ""



import dataclasses
import pathlib
import typing

from p2g import gbl


@dataclasses.dataclass
class NodePlace:
    col_offset_first: int
    col_offset_past: int
    lineno: int
    funcname: str
    filename: str


@dataclasses.dataclass
class State:
    last_pos: NodePlace

    def __init__(self):
        self.last_pos = NodePlace(0, 0, 0, "", "")


state = State()


def mark_pos(last_pos):
    state.last_pos = last_pos


def src_code_from_line_near(pos, lineno):
    with gbl.openr(pos.filename) as (inf, etrace):
        if etrace:
            return ""
        src = gbl.logread(inf)
        lines = gbl.splitnl(src)
        return lines[lineno - 1]


def src_code_from_node_place(pos):
    return src_code_from_line_near(pos, pos.lineno)


def note_from_node_place(pos, absolute_lines, topfn=None):
    relfixer = 1
    if (gbl.config.tin_test or (not absolute_lines)) and topfn is not None:
        relfixer = topfn.__code__.co_firstlineno

    filename = pos.filename
    if gbl.config.tin_test:
        orig = pathlib.Path(filename)
        filename = "/".join(orig.parts[-3:])

    reportable_line = pos.lineno - relfixer + 1
    col_offset = pos.col_offset_first
    col_width = pos.col_offset_past - col_offset
    pfx = ":".join(
        [
            str(filename),
            str(reportable_line),
            str(pos.col_offset_first),
            str(pos.col_offset_past),
            " ",
        ]
    )

    return pfx, " " * (len(pfx) + col_offset) + "^" * col_width


class CompilerError(Exception):
    pos: NodePlace
    message: str

    # def __str__(self):
    #     return str(self.message)

    def __init__(self, pos, message: str):
        super().__init__()
        self.pos = pos
        self.message = message

    def error_lines(self, absolute_lines, topfn):
        sourceline = src_code_from_node_place(self.pos)
        pfx, carat = note_from_node_place(self.pos, absolute_lines, topfn)

        return [self.message, pfx + sourceline, carat]

    def report_error(self, absolute_lines=True, topfn=None):
        for line in self.error_lines(absolute_lines, topfn):
            gbl.eprint(line)


def compiler(message, exc=None, err_pos=None) -> typing.NoReturn:
    if gbl.config.bp_on_error:  # for debug
        breakpoint()
        breakpoint()

    if not err_pos:
        err_pos = state.last_pos
    raise CompilerError(err_pos, str(message)) from exc



import abc
import ast
import dataclasses
import pathlib
import re
import sys

from p2g import axis
from p2g import coords
from p2g import err
from p2g import gbl
from p2g import nd
from p2g import op
from p2g import scalar
from p2g import stat
from p2g import symbol
from p2g import vector
from p2g import walkbase
from p2g import walkexpr
from p2g import walkfunc


# wrap enough state that we can call an
# ast function as if it were native.


def ensure_no_more(src_gen):
    try:
        next(src_gen)
        err.compiler("Too many values to unpack.")

    except StopIteration:
        pass


def handle_assign(self, target, val):
    if not isinstance(target, ast.Tuple):
        self.visit_store(target, val)
        return

    src_gen = iter(val)
    try:
        for dst_idx, dst_el in enumerate(target.elts):
            if isinstance(dst_el, ast.Starred):
                src_togo = list(src_gen)
                break_i = len(src_togo) - (len(target.elts) - dst_idx - 1)
                self.visit_store(dst_el.value, src_togo[:break_i])
                src_gen = iter(src_togo[break_i:])
            else:
                self.visit_store(dst_el, next(src_gen))

    except StopIteration as exc:
        err.compiler("Not enough values to unpack.", exc)

    ensure_no_more(src_gen)


# make code for while and for loops.


class Itr(abc.ABC):
    @abc.abstractmethod
    def __init__(self, itr, target):
        pass

    @abc.abstractmethod
    def cur_index(self):
        pass

    @abc.abstractmethod
    def last_index(self):
        pass

    @abc.abstractmethod
    def next_target(self):
        pass

    @abc.abstractmethod
    def next_index(self):
        pass


@dataclasses.dataclass
class ItrRange:
    target: ast.AST

    def __init__(self, itr, interp, target):
        self.interp = interp
        self.rang = itr
        self.var = coords.Var()

        if not isinstance(target, ast.Name):
            err.compiler("Must be simple name as destination for 'for'.")
        self.var = coords.Var()
        interp.ns[target.id] = self.var

    #        interp.ns[target.id] = self.var

    def setup_index(self):
        pass

    def setup_target(self):
        stat.add_stat(
            stat.Set(
                scalar.wrap_scalar(self.var),
                scalar.wrap_scalar(self.rang.start),
            )
        )

    #        breakpoint()
    #        handle_assign(self.interp, self.var, self.rang.start)

    def cur_index(self):
        return self.var

    def past_index(self):
        return self.rang.stop

    def load_target(self):
        pass

    def next_index(self):
        # stat.add_stat(
        #     stat.Set(
        #         scalar.wrap_scalar(self.target),
        #         op.make_scalar_add(scalar.wrap_scalar(self.target), self.step),
        #     )
        # )

        stat.add_stat(
            stat.Set(
                scalar.wrap_scalar(self.var),
                op.make_scalar_add(scalar.wrap_scalar(self.var), self.rang.step),
            )
        )


@dataclasses.dataclass
class ItrSlice:
    ptr: nd.EBase
    pastptr: scalar.Scalar
    step: scalar.Scalar
    var: ast.AST

    def __init__(self, itr, interp, target):
        self.interp = interp
        self.target = target

        # with a slice, always need a pointer to
        # look at things.
        self.ptr = coords.Var(itr._addr)  # scalar.wrap_scalar(coords.Var(itr._addr))

        if not isinstance(target, ast.Name):
            err.compiler("Must be simple name as destination for 'for'.")

        self.pastptr = itr._addr + itr._size
        self.step = itr._step

    def setup_index(self):
        pass

    def setup_target(self):
        pass

    def cur_index(self):
        return self.ptr

    def past_index(self):
        return self.pastptr

    def load_target(self):
        handle_assign(self.interp, self.target, op.hashop(scalar.wrap_scalar(self.ptr)))

    #        stat.add_stat(stat.Set(self.interp.visit(self.var), op.hashop(self.ptr)))

    def next_index(self):
        stat.add_stat(
            stat.Set(
                scalar.wrap_scalar(self.ptr),
                op.make_scalar_add(scalar.wrap_scalar(self.ptr), self.step),
            )
        )


def make_itr(itr, interp, target):
    if isinstance(itr, vector.Vec):
        return ItrSlice(itr, interp, target)
    if isinstance(itr, range):
        return ItrRange(itr, interp, target)

    err.compiler("Illegal iterator.")


def geniter(
    self,
    *,
    whileexp=None,
    target=None,
    itr=None,
    body=None,
    orelse=None,
):
    self.loop = walkbase.LoopContext(self.loop)

    if whileexp:
        stat.add_stat(stat.LabelDef(self.loop.lcontinue))
        test = op.make_scalar_unop(op.a2opfo(ast.Not), self.visit(whileexp))
        stat.add_stat(stat.If(test, self.loop.lorelse))
        self.visit_slist(body)
        stat.add_stat(stat.Goto(self.loop.lcontinue))
    else:
        assert itr

        itr = make_itr(itr, self, target)

        itr.setup_index()
        itr.setup_target()
        stat.add_stat(stat.LabelDef(self.loop.lcontinue))

        stat.add_stat(
            stat.If(
                op.make_vec_binop(
                    op.a2opfo(ast.GtE),
                    itr.cur_index(),
                    itr.past_index(),
                ),
                self.loop.lorelse,
            )
        )
        itr.load_target()
        self.visit_slist(body)
        itr.next_index()
        stat.add_stat(stat.Goto(self.loop.lcontinue))

    if orelse:
        stat.add_stat(stat.LabelDef(self.loop.lorelse))
        self.visit_slist(orelse)
        if self.loop.lbreak.used:
            stat.add_stat(stat.LabelDef(self.loop.lbreak))
    else:
        if self.loop.lbreak.used:
            stat.add_stat(stat.LabelDef(self.loop.lbreak))
        stat.add_stat(stat.LabelDef(self.loop.lorelse))
    self.loop = self.loop.prev


class WalkStatement(walkbase.WalkBase):
    def wrap_decorators(self, obj, node):
        for deco_n in reversed(node.decorator_list):
            deco = self.visit(deco_n)
            obj = deco(obj)
        return obj

    def _visit_module(self, node, file_name):
        self.ns = self.module_ns = walkbase.ModuleNS()
        self.ns["__file__"] = file_name
        self.file_name = file_name
        self.ns["__name__"] = "__main__"

        self.visit_slist(node.body)

    def visit_module(self, node, file_name):
        self._visit_module(node, file_name)

    def _visit_classdef(self, node):
        clsns = walkbase.ClassNS()
        with self.pushpopns(clsns):
            self.visit_slist(node.body)

        cls = type(
            node.name,
            tuple(self.visit(b) for b in node.bases),
            clsns.guts,
        )
        cls = self.wrap_decorators(cls, node)
        self.ns[node.name] = cls

    def _visit_for(self, node):
        itr = self.visit(node.iter)
        return geniter(
            self, target=node.target, itr=itr, body=node.body, orelse=node.orelse
        )

    def _visit_while(self, node):
        geniter(self, whileexp=node.test, body=node.body, orelse=node.orelse)

    def _visit_break(self, _):
        self.loop.lbreak.used = True
        stat.add_stat(stat.Goto(self.loop.lbreak))

    def _visit_continue(self, _):
        stat.add_stat(stat.Goto(self.loop.lcontinue))

    def _visit_if(self, node):
        #   for nicer looking code
        # for a conditional break.
        if isinstance(node.body[0], ast.Break):
            exp = op.make_scalar_unop(op.a2opfo(ast.UAdd), self.visit(node.test))
            self.loop.lbreak.used = True
            stat.add_stat(stat.If(exp, self.loop.lbreak))
            return

        elsepart = stat.next_label()
        donepart = stat.next_label()

        cond = op.make_scalar_unop(op.a2opfo(ast.UAdd), self.visit(node.test))

        # nice code for if with constant.
        if cond.is_constant:
            if cond:
                self.visit_slist(node.body)
            else:
                self.visit_slist(node.orelse)
        else:
            exp1 = op.make_scalar_unop(op.a2opfo(ast.Not), cond)
            stat.add_stat(stat.If(exp1, on_t=elsepart))
            self.visit_slist(node.body)
            stat.add_stat(stat.Goto(donepart))
            stat.add_stat(stat.LabelDef(elsepart))
            self.visit_slist(node.orelse)
            stat.add_stat(stat.LabelDef(donepart))

    def _visit_import(self, node):
        for n in node.names:
            self.ns[n.asname or n.name] = __import__(n.name)

    def _visit_importfrom(self, node):
        mod = __import__(
            node.module, None, None, [n.name for n in node.names], node.level
        )

        for n in node.names:
            if n.name == "*":
                for modname, modguts in mod.__dict__.items():
                    if modname[0] != "_":
                        self.ns[modname] = modguts
            else:
                self.ns[n.asname or n.name] = getattr(mod, n.name)

    def _visit_augassign(self, node):
        self.visit_store(
            node.target,
            op.make_vec_binop(
                op.a2opfo(type(node.op)),
                self.visit(node.target),
                self.visit(node.value),
            ),
        )

    def _visit_annassign(self, node):
        if node.value is not None:
            val = self.visit(node.value)
            handle_assign(self, node.target, val)

    def _visit_assign(self, node):
        val = self.visit(node.value)
        for n in node.targets:
            handle_assign(self, n, val)

    def _visit_delete(self, node):
        for n in node.targets:
            self.visit(n)

    def _visit_pass(self, _):
        pass

    def _visit_assert(self, node):
        res = self.visit(node.test)
        if node.msg is None:
            if not res:
                assert res
        else:
            assert res, self.visit(node.msg)


class Walk(
    WalkStatement,
    walkexpr.WalkExpr,
    walkbase.WalkNS,
    walkfunc.WalkFunc,
    walkbase.WalkBase,
):
    pass


def compile_all(node, srcfile_name):
    walker = Walk()
    walker.visit_module(node, srcfile_name)
    return walker


def find_defined_funcs(sourcelines):
    for line in sourcelines.split("\n"):
        # find last line with def in it, that's the function we need
        mares = re.match("def (.*?)\\(", line)
        if mares:
            yield mares.group(1)


def find_main_func_name(sourcelines, func_name_arg):
    if func_name_arg != "<last>":
        return func_name_arg
    function_to_call = "no function in file"
    for fname in find_defined_funcs(sourcelines):
        function_to_call = fname
    return function_to_call


@gbl.g2l
def compile2g(func_name_arg, srcfile_name, job_name, args=None):
    srcpath = pathlib.Path(srcfile_name)

    with gbl.openr(srcpath) as (inf, etrace):
        if etrace:
            err.compiler(f"File '{srcpath}' not found.")
        sys.path.insert(0, str(srcpath.parent))

        with stat.Nest() as cursor:
            axis.NAMES = "xyz"
            gbl.iface.reset()

            symbol.Table.reset()

            gbl.log(f"Starting {func_name_arg} {cursor.next_label}")
            sourcelines = gbl.logread(inf)
            node = ast.parse(sourcelines)
            # load everything
            walker = compile_all(node, srcfile_name)

            if node.body:
                walkfunc.digest_top(
                    walker,
                    find_main_func_name(sourcelines, func_name_arg),
                    srcpath,
                    job_name,
                    args,
                )
            # can't use generator 'cause need
            # mods to varrefs for symbol table
            res = list(cursor.to_full_lines())

            yield from symbol.Table.yield_lines()
            yield from res



import ast
import dataclasses
import itertools
import typing

from p2g import err
from p2g import gbl
from p2g import op
from p2g import stat
from p2g import walkbase


class Marker:
    pass


# look through caller's args and insert defaults
# etc. to be compatible with callee, return dict
# with mapping.


def formal_kwargs(formalspec, formals, kwargs, pos) -> dict[str, typing.Any]:
    # Process incoming keyword arguments, putting them in namespace if
    # actual arg exists by that name, or offload to function's kwarg
    # if any. All make needed checks and error out.
    func_kwarg = {}
    final_dict = {}
    all_formals = {
        x.arg
        for x in itertools.chain(formals.args, formals.kwonlyargs, formals.kw_defaults)
        if x is not None
    }

    for key, value in kwargs.items():
        if key in all_formals:
            final_dict[key] = value
        elif formalspec.kwarg:
            func_kwarg[key] = value
        else:
            err.compiler("Bad arguments.", err_pos=pos)

    if formalspec.kwarg:
        final_dict[formalspec.kwarg.arg] = func_kwarg
    return final_dict


def check_missing(final_dict, formalspec, pos):
    # check for missing args
    for formal in itertools.chain(formalspec.args, formalspec.kwonlyargs):
        if formal.arg not in final_dict:
            err.compiler(f"Missing argument '{formal.arg}'.", err_pos=pos)


def get_defaults(walker, formals):
    first_defaulted = len(formals.args) - len(formals.defaults)
    return {
        # position args defaults
        el.arg: walker.visit(formals.defaults[idx - first_defaulted])
        for idx, el in enumerate(formals.args)
        if idx >= first_defaulted
    } | {
        # kwargs defaults
        el.arg: walker.visit(val)
        for el, val in zip(formals.kwonlyargs, formals.kw_defaults)
        if val is not None
    }


def gather_func_formals(func_def, *args, **kwargs):
    walker = func_def.walker
    formals = func_def.node.args

    # report error in caller rather than definition.
    pos = err.state.last_pos

    final_dict = {}

    formalspec = func_def.node.args

    if formalspec.vararg:
        final_dict[formalspec.vararg.arg] = args[len(formalspec.args) :]
    else:
        if len(args) > len(formalspec.args):
            err.compiler("Bad arguments.", err_pos=pos)

    for i in range(min(len(args), len(formalspec.args))):
        final_dict[formalspec.args[i].arg] = args[i]

    # result contains at least the defaults.
    final_dict |= get_defaults(walker, formals)
    final_dict |= formal_kwargs(formalspec, formals, kwargs, pos)

    check_missing(final_dict, formalspec, pos)

    err.state.last_pos = pos
    return final_dict


# calling a function, emit code to call, and
# generate the called code.  And save it, to make
# sure that other generations make the same.


def inline(func_def, *args, **kwargs):
    walker = func_def.walker
    prev_file = walker.file_name
    prev_func = walker.func_name

    walker.file_name = func_def.file_name
    walker.func_name = func_def.func_name

    # We need to switch from dynamic execution scope to lexical scope
    # in which function was defined (then switch back on return).
    dyna_scope = walker.ns
    walker.ns = func_def.lexical_scope
    res = None

    with walker.pushpopns(walkbase.FunctionNS()):
        formals_dict = gather_func_formals(func_def, *args, **kwargs)

        walker.ns.guts.update(formals_dict)

        res = walker.visit_slist(func_def.node.body)

    walker.ns = dyna_scope
    walker.func_name = prev_func
    walker.file_name = prev_file
    return res


@dataclasses.dataclass
class FuncDefWrap:
    file_name: str
    func_name: str

    node: ast.AST
    walker: "WalkFunc"
    lexical_scope: walkbase.ANamespace
    call: bool

    gen: list[stat.StatBase]

    def __init__(self, walker, node):
        self.call = False
        self.gen = []
        self.func_name = node.name

        self.file_name = walker.module_ns["__file__"]

        self.node = node
        self.walker = walker
        self.lexical_scope = walker.ns

    def __call__(self, *args, **kwargs):
        return inline(self, *args, **kwargs)


def interpfunc(fun):
    def func(*args, **kwargs):
        if args and isinstance(args[0], Marker):
            return fun
        return fun(*args, **kwargs)

    return func


@dataclasses.dataclass
class FuncArgsDescr:
    args: list[typing.Any]
    kwargs: dict[str, typing.Any]

    def __init__(self, walker, node):
        args = []
        for arg in node.args:
            if isinstance(arg, ast.Starred):
                args.extend(walker.visit(arg.value))
            else:
                args.append(walker.visit(arg))

        kwargs = {}
        for keyword in node.keywords:
            val = walker.visit(keyword.value)
            if keyword.arg is None:
                kwargs.update(val)
            else:
                kwargs[keyword.arg] = val
        self.args = args
        self.kwargs = kwargs


def funcall(target, *args, **kwargs):
    return target(*args, **kwargs)


class WalkFunc(walkbase.WalkBase):
    def _visit_call(self, node):
        defn = self.visit(node.func)
        desc = FuncArgsDescr(self, node)
        if defn.__module__ == "p2g.builtin":
            return op.make_scalar_func(defn.__name__, *desc.args)

        #        return desc.callit(*desc.args, **desc.kwargs)
        # f = interpfunc(desc.func)
        # return f
        return funcall(defn, *desc.args, **desc.kwargs)

    def _visit_functiondef(self, node):
        desc = FuncDefWrap(self, node)
        #        self.funcmaps[node.name] = desc
        ifunc = interpfunc(desc)
        self.ns[node.name] = ifunc


def digest_top(walker, func_name, srcpath, job_name, args):
    if args is None:
        args = []
    try:
        fncdef = walker.ns[func_name]
        desc = fncdef(Marker())
    except KeyError:
        err.compiler(f"No such function '{func_name}' in '{srcpath}'.")

    def insides():
        if gbl.config.bp_on_error:  # no cover
            inline(desc, *args)
        else:
            try:
                inline(desc, *args)
            except (AttributeError, IndexError) as exn:
                err.compiler(exn)

    if gbl.config.boiler_plate and job_name:
        stat.code([job_name], comment_txt=func_name)
        insides()
        stat.add_stat(stat.Code("M30", "<none>"))
    else:
        insides()



from p2g import err
from p2g import gbl
from p2g import stat
from p2g import vector


def set_wcs(new_wcs: vector.Vec):
    gcode = (new_wcs.get_address() - 5221) // 20 + 54
    stat.code(f"G{gcode}")


def base_addr(new_base=None):
    #    ._wcs = (val._addr - 5221) // 20 + 54
    if new_base is not None:
        gbl.iface.ebss = new_base
    return gbl.iface.ebss


def message(dst, txt: str):
    stat.code(f"{dst[0]} = 101 ({txt})", comment_txt="<none>")


def as_address(src):
    return src


def address(src):
    # special case for vecs, give address
    # anyway.
    src = vector.wrap_maybe_vec(src)
    add = src.get_address()
    if add is None:
        err.compiler("Can only take address of something with location.")
    return add


def alias(thing):
    return thing



import abc
import ast
import math
import typing

from p2g import err
from p2g import lib
from p2g import nd
from p2g import scalar
from p2g import symbol
from p2g import vector


OptRes = typing.Optional[scalar.Scalar]
ScalarScalar = typing.Union[scalar.Scalar, int, float, bool]

op_byclass: typing.Dict[typing.Any, nd.Opinfo] = {}
allops: typing.Dict[str, nd.Opinfo] = {}


def parensif(cond, thing):
    if cond:
        return "[" + thing + "]"
    return thing


def get_prec(thing):
    return getattr(thing, "prec", 20)


class Binop(scalar.Scalar):
    def __init__(self, opfo, lhs, rhs):
        super().__init__(opfo)
        assert opfo.pyn != "-rev"
        self.lhs = lhs
        self.rhs = rhs

    def same(self, other):
        return (
            self.opfo == other.opfo
            and lib.same(self.lhs, other.lhs)
            and lib.same(self.rhs, other.rhs)
        )

    def to_gcode(self, modi: nd.NodeModifier) -> str:
        if self.opfo.g_func:
            res = [
                nd.to_gcode(self.lhs, modi),
                nd.to_gcode(self.rhs, modi),
            ]
            return f"{self.opfo.gname}[{','.join(res)}]"

        res = []

        outer_prec = self.opfo.prec

        def handle_term(node):
            return parensif(
                get_prec(node) < outer_prec,
                nd.to_gcode(node, modi),
            )

        res.append(handle_term(self.lhs))
        res.append(self.opfo.gname)
        res.append(handle_term(self.rhs))

        if modi & nd.NodeModifier.NOSPACE:
            return "".join(res)
        return " ".join(res)

    def __repr__(self):
        return f"({self.lhs}{self.opfo.pyn}{self.rhs})"


def make_scalar_binop(opfo, lhs, rhs):
    # can it be done right now?
    if isinstance(rhs, scalar.ConstantNone):
        rhs = scalar.wrap_scalar(0)
    if isinstance(lhs, scalar.Constant) and isinstance(rhs, scalar.Constant):
        lval = lhs.value
        rval = rhs.value
        if isinstance(lval, float) or isinstance(rval, float):
            lval = float(lval)
            rval = float(rval)
        if opfo.pyn == "round":
            rval = int(rval)
        res = getattr(lval, opfo.mth)(rval)
        return scalar.Constant(res)
    # put contant on rhs if we can
    if (
        opfo.comm
        and isinstance(lhs, scalar.Constant)
        and not isinstance(rhs, scalar.Constant)
    ):
        lhs, rhs = rhs, lhs

    if opfo.opt2 and (res := opfo.opt2(opfo, lhs, rhs)) is not None:
        return res

    return Binop(opfo, lhs, rhs)


def make_scalar_add(lhs, rhs):
    return make_scalar_binop(allops["+"], lhs, rhs)


def a2opfo(optyp):
    return op_byclass[optyp]


def opt1_func(opfo, arg) -> OptRes:
    if not isinstance(arg, scalar.Constant):
        return None

    return scalar.wrap_scalar(opfo.lam(arg.to_float()))


class Unop(scalar.Scalar):
    def __init__(self, opfo, child):
        super().__init__(opfo)
        self.child = child

    def get_address(self):
        return self.child

    def same(self, other):
        return self.opfo == other.opfo and lib.same(self.child, other.child)

    def to_gcode(self, modifier=nd.NodeModifier.EMPTY):
        res = []

        if self.opfo.pyn == "#":
            try:
                addr = int(self.child)
                symbol.Table.add_varref(addr, err.state.last_pos)
            except TypeError:
                pass

        outer_prec = self.opfo.prec

        res.append(self.opfo.gname)
        # at least # is left right associative.
        inside_prec = get_prec(self.child)

        if self.opfo.pyn == "#":
            modifier |= nd.NodeModifier.ADDRESS

        res.append(
            parensif(
                outer_prec >= inside_prec,
                nd.to_gcode(self.child, modifier),
            )
        )

        return "".join(res)

    def __repr__(self):
        return f"({self.opfo.pyn} {self.child})"


def make_scalar_unop(opfo, child: ScalarScalar):
    child = scalar.wrap_scalar(child)
    if opfo.opt1 and (res := opfo.opt1(opfo, child)) is not None:
        return res

    if isinstance(child, scalar.Constant):
        return scalar.wrap_scalar(getattr(child.value, opfo.mth)())

    return Unop(opfo, child)


def make_vec_binop(opfo, lhs, rhs=None, force_ourtype=False):
    if not force_ourtype and not isinstance(lhs, (int, float, nd.EBase)):
        return getattr(lhs, opfo.mth)(rhs)

    lhs = vector.wrap_maybe_vec(lhs)

    if rhs is None:
        # actually unop.

        return vector.sorv_from_list(
            [make_scalar_unop(opfo, el) for el in lhs.everything()]
        )
    rhs = vector.wrap_maybe_vec(rhs)
    return vector.sorv_from_list(
        [
            make_scalar_binop(opfo, lel, rel)
            for lel, rel in zip(lhs.everything(), rhs.forever())
        ]
    )


def make_scalar_func(fname, *args):
    ofo = allops[fname]
    return make_vec_binop(ofo, args[0], force_ourtype=True)


class MAST(abc.ABC):
    opfo: nd.Opinfo


MAST.register(ast.AST)


class HASH(MAST):
    pass


class ABS(MAST):
    pass


class RSUB(MAST):
    pass


class ROUND(MAST):
    pass


revop = {
    "+": "+",
    "-": "+",
    "*": "*",
    "/": "*",
}


# a  + k1  + k2 -> a + (k1+k2)
def opt2_fold(opfo, lhs, rhs) -> OptRes:
    rop = revop.get(opfo.pyn)

    if (
        rop
        and isinstance(lhs, Binop)
        and lhs.opfo == opfo
        and lhs.rhs.is_constant
        and rhs.is_constant
    ):
        return make_scalar_binop(
            opfo,
            lhs.lhs,
            make_scalar_binop(
                allops[rop],
                lhs.rhs,
                rhs,
            ),
        )

    return None


def opt2_mul(opfo, lhs, rhs) -> OptRes:
    match rhs.to_float():
        case 0.0:
            return scalar.Constant(0.0)
        case 1.0:
            return lhs
        case -1.0:
            return make_scalar_unop(allops["un-"], lhs)

    return opt2_fold(opfo, lhs, rhs)


def opt2_div(opfo, lhs, rhs) -> OptRes:
    match rhs.to_float():
        case None:
            return None
        case 0.0:
            err.compiler("Attempt to divide by zero.")
        case 1.0:
            return lhs
        case -1.0:
            return make_scalar_unop(allops["un-"], lhs)

    return opt2_fold(opfo, lhs, rhs)


not_compop = {
    "==": "!=",
    "!=": "==",
    "<": ">=",
    ">": "<=",
    "<=": ">",
    ">=": "<",
}


def opt1_not(_nd, arg) -> OptRes:
    if arg.is_constant:
        return scalar.Constant(not arg.value)

    # got not(comop, turn into (inverted compop))

    if notted := not_compop.get(arg.opfo.pyn):
        return make_scalar_binop(allops[notted], arg.lhs, arg.rhs)

    rhs = scalar.Constant(0)
    return make_scalar_binop(allops["!="], arg, rhs)


def opt1_plus(_, arg) -> OptRes:
    return arg


# don't have binary not operator, make from xor.
def opt1_invert(_, arg) -> OptRes:
    rhs = scalar.wrap_scalar(-1)
    return make_scalar_binop(allops["^"], arg, rhs)


# turn a1 > a2  == 0
# => (a1 <= a2)


def opt2_eq(opfo, lhs: scalar.Scalar, rhs: scalar.Scalar) -> OptRes:
    if lib.same(lhs, rhs):
        return scalar.Constant(opfo.pyn == "==")

    if not isinstance(rhs, scalar.Constant) or rhs.value != 0.0:
        return None

    if isinstance(lhs, Binop):
        if rev := not_compop.get(lhs.opfo.pyn):
            if opfo.pyn == "==":
                return make_scalar_binop(
                    allops[rev],
                    lhs.lhs,
                    lhs.rhs,
                )
            return lhs
    return None


def opt2_matmul(_opfo, _lhs, _rhs) -> OptRes:  # for debug
    return None


def opt2_add(opfo, lhs, rhs) -> OptRes:
    if (res := opt2_fold(opfo, lhs, rhs)) is not None:
        return res

    if isinstance(rhs, Unop) and rhs.opfo.pyn == "un-":
        return make_scalar_binop(allops["-"], lhs, rhs.child)

    if isinstance(rhs, scalar.Constant):
        rhsv = rhs.to_float()
        if rhsv < 0.0:
            return make_scalar_binop(allops["-"], lhs, -rhsv)
        if rhsv == 0.0:
            return lhs

    return None


def make_fmt(val, fmt):
    if isinstance(val, str):
        return val
    val = scalar.wrap_scalar(val)
    if isinstance(val, scalar.Constant):
        if fmt and fmt[-1] == "x":
            return format(val.to_int(), fmt)
        return format(val.to_float(), fmt)
    gfmt = f"[{fmt[0]}{fmt[2]}]" if fmt else ""

    return "[" + val.to_gcode(nd.NodeModifier.NOSPACE) + "]" + gfmt


def opt2_sub(opfo, lhs, rhs) -> OptRes:
    if (res := opt2_fold(opfo, lhs, rhs)) is not None:
        return res

    # a -  - b -> a + b
    if isinstance(rhs, Unop) and rhs.opfo.pyn == "un-":
        return make_scalar_binop(allops["+"], lhs, rhs.child)

    if isinstance(rhs, scalar.Constant):
        rhsv = rhs.to_float()
        if rhsv == 0.0:
            return lhs

    return None


def unwrap_int(val):
    try:
        return val.to_int()
    except AttributeError:
        return val


def make_slice(low, high, step):
    return slice(
        unwrap_int(low),
        unwrap_int(high),
        unwrap_int(step),
    )


def hashop(arg):
    return Unop(allops["#"], arg)


def make_hashop(lhs, rhs):
    return hashop(make_scalar_add(lhs, rhs))


vector.MemVec.make_hashop = make_hashop


def nd_install(opfo: nd.Opinfo):
    def make_multi_binop_tramp(lhs, rhs):
        return make_vec_binop(opfo, lhs, rhs)

    def make_multi_unop_tramp(child):
        return vector.sorv_from_list(
            [make_scalar_unop(opfo, cel) for cel in child.everything()]
        )

    def make_scalar_binop_tramp(lhs, rhs):
        return make_scalar_binop(opfo, lhs, scalar.wrap_scalar(rhs))

    def make_scalar_unop_tramp(child):
        return make_scalar_unop(opfo, child)

    if opfo.mth:
        if isinstance(opfo.mth, str):
            if opfo.nargs == 1:
                setattr(vector.Vec, opfo.mth, make_multi_unop_tramp)
                setattr(scalar.Scalar, opfo.mth, make_scalar_unop_tramp)
            else:
                setattr(vector.Vec, opfo.mth, make_multi_binop_tramp)
                setattr(scalar.Scalar, opfo.mth, make_scalar_binop_tramp)


def reg(**kwargs):
    opi = nd.Opinfo(**kwargs)
    allops[opi.pyn] = opi
    op_byclass[opi.astc] = opi
    nd_install(opi)


def regfunc(name, lam):
    reg(
        pyn=name,
        lam=lam,
        gname=name.upper(),
        prec=20,
        nargs=1,
        opt1=opt1_func,
    )


regfunc("cos", lambda x: math.cos(math.radians(x)))
regfunc("sin", lambda x: math.sin(math.radians(x)))
regfunc("tan", lambda x: math.tan(math.radians(x)))
regfunc("acos", lambda x: math.degrees(math.acos(x)))
regfunc("asin", lambda x: math.degrees(math.asin(x)))
regfunc("atan", lambda x: math.degrees(math.atan(x)))
regfunc("exp", math.exp)
regfunc("ln", math.log)
regfunc("sqrt", math.sqrt)
regfunc("exists", id)
regfunc("fup", math.ceil)
regfunc("fix", math.floor)
regfunc("ground", round)


# fmt: off
# noqa: E203
reg(astc=ast.BitOr    , pyn="|"     , mth="__or__"       , gname="OR"    , prec=4, comm=True)
reg(astc=ast.BitXor   , pyn="^"     , mth="__xor__"      , gname="XOR"   , prec=5, comm=True)
reg(astc=ast.BitAnd   , pyn="&"     , mth="__and__"      , gname="AND"   , prec=6, comm=True)
reg(astc=ast.Lt       , pyn="<"     , mth="__lt__"       , gname="LT"    , prec=10)
reg(astc=ast.LtE      , pyn="<="    , mth="__le__"       , gname="LE"    , prec=10)
reg(astc=ast.NotEq    , pyn="!="    , mth="__ne__"       , gname="NE"    , prec=10, comm=True, opt2=opt2_eq)
reg(astc=ast.Eq       , pyn="=="    , mth="__eq__"       , gname="EQ"    , prec=10, comm=True, opt2=opt2_eq)
reg(astc=ast.Gt       , pyn=">"     , mth="__gt__"       , gname="GT"    , prec=10)
reg(astc=ast.GtE      , pyn=">="    , mth="__ge__"       , gname="GE"    , prec=10)
reg(astc=ast.Sub      , pyn="-"     , mth="__sub__"      , gname="-"     , prec=12, opt2=opt2_sub)
reg(astc=None         , pyn="-rev"  , mth="__rsub__"     , gname="-rev"  , prec=12)
reg(astc=ast.Add      , pyn="+"     , mth="__add__"      , gname="+"     , prec=12, comm=True, opt2=opt2_add)
reg(astc=None         , pyn="round" , mth="__round__"    , gname="round" , prec=12)
reg(astc=ast.Mult     , pyn="*"     , mth="__mul__"      , gname="*"     , prec=13, comm=True, opt2=opt2_mul)
reg(astc=ast.Div      , pyn="/"     , mth="__truediv__"  , gname="/"     , prec=13, opt2=opt2_div)
reg(astc=ast.FloorDiv , pyn="//"    , mth="__floordiv__" , gname="//"    , prec=13, opt2=opt2_div)
reg(astc=ast.Mod      , pyn="%"     , mth="__mod__"      , gname="MOD"   , prec=14)
reg(astc=ast.Pow      , pyn="**"    , mth="__pow__"      , gname="POW"   , prec=15, g_func=True)
reg(astc=ast.USub     , pyn="un-"   , mth="__neg__"      , gname="-"     , prec=14, nargs=1)
reg(astc=ast.UAdd     , pyn="un+"   , mth="__pos__"      , gname="+"     , prec=14, nargs=1, opt1=opt1_plus)
reg(astc=ast.MatMult  , pyn="m*"    , mth=""             , gname="ERRR"  , prec=20, nargs=2, opt2=opt2_matmul)
reg(astc=ast.Invert   , pyn="un~"   , mth="__invert__"   , gname="un~"   , prec=14, nargs=1, opt1=opt1_invert)
reg(astc=ast.Not      , pyn="unot"  , mth=""             , gname="unot"  , prec=14, nargs=1, opt1=opt1_not)
reg(astc=None         , pyn="abs"   , mth="__abs__"      , gname="ABS"   , prec=15, nargs=1, g_func=True)
reg(astc=None         , pyn="#"     , mth=""             , gname="#"     , prec=19, nargs=1)


# fmt:on



from .axis import NAMES
from .builtin import abs
from .builtin import acos
from .builtin import asin
from .builtin import atan
from .builtin import cos
from .builtin import exists
from .builtin import exp
from .builtin import fix
from .builtin import fup
from .builtin import ground
from .builtin import ln
from .builtin import sin
from .builtin import sqrt
from .builtin import tan
from .coords import Const
from .coords import Fixed
from .coords import Var
from .goto import goto

# from .ptest import must_be

from .scalar import Scalar
from .stat import code
from .stat import codenl
from .stat import com
from .stat import comment
from .stat import dprint
from .symbol import Table
from .vector import RValueVec
from .vector import Vec
from .visible import address
from .visible import alias
from .visible import as_address
from .visible import base_addr
from .visible import message
from .visible import set_wcs

__all__ = [
    "NAMES",
    "abs",
    "acos",
    "asin",
    "atan",
    "cos",
    "exists",
    "exp",
    "fix",
    "fup",
    "ground",
    "ln",
    "sin",
    "sqrt",
    "tan",
    "Const",
    "Fixed",
    "Var",
    "goto",
    "Scalar",
    "code",
    "codenl",
    "com",
    "comment",
    "dprint",
    "Table",
    "RValueVec",
    "Vec",
    "address",
    "alias",
    "as_address",
    "base_addr",
    "message",
    "set_wcs",
]


VERSION = "0.2.29"



import p2g

from p2g import axis


NAME = "HAAS NGC"
ORIGIN = [0, 0, 0]
NAXES = axis.Axes()

NO_SKIP = "M78"
MUST_SKIP = "M79"
FEED_FAST = 650.0
FEED_SLOW = 50.0
PROBE_FAST = 50.0
PROBE_SLOW = 10.0
PROBE_ON = ["G65 P9832"]
PROBE_H = p2g.Fixed[3](addr=2001)
OTS_ON = ["M59 P2", "G04 P1.0", "M59 P3"]
OTS_OFF = ["M69 P2", "M68 P3"]
SPINDLE_PROBE_ON = "G65 P9832"
SPINDLE_PROBE_OFF = "G65 P9833"
NO_LOOKAHEAD = ["G103 P1", "G04 P1", "G04 P1", "G04 P1"]

# MACHINE GEN BELOW
NULL = p2g.Fixed(addr=0)
MACRO_ARGUMENTS = p2g.Fixed[33](addr=1)
# 34 .. 99 - ..
GP_SAVED1 = p2g.Fixed[100](addr=100)
# 200 .. 499 - ..
GP_SAVED2 = p2g.Fixed[50](addr=500)
PROBE_CALIBRATION1 = p2g.Fixed[6](addr=550)
PROBE_R = p2g.Fixed[3](addr=556)
PROBE_CALIBRATION2 = p2g.Fixed[22](addr=559)
GP_SAVED3 = p2g.Fixed[119](addr=581)
# 700 .. 799 - ..
GP_SAVED4 = p2g.Fixed[200](addr=800)
INPUTS = p2g.Fixed[64](addr=1000)
MAX_LOADS_XYZAB = p2g.Fixed[5](addr=1064)
# 1069 .. 1079 - ..
RAW_ANALOG = p2g.Fixed[10](addr=1080)
FILTERED_ANALOG = p2g.Fixed[8](addr=1090)
SPINDLE_LOAD = p2g.Fixed(addr=1098)
# 1099 .. 1263 - ..
MAX_LOADS_CTUVW = p2g.Fixed[5](addr=1264)
# 1269 .. 1600 - ..
TOOL_TBL_FLUTES = p2g.Fixed[200](addr=1601)
TOOL_TBL_VIBRATION = p2g.Fixed[200](addr=1801)
TOOL_TBL_OFFSETS = p2g.Fixed[200](addr=2001)
TOOL_TBL_WEAR = p2g.Fixed[200](addr=2201)
TOOL_TBL_DROFFSET = p2g.Fixed[200](addr=2401)
TOOL_TBL_DRWEAR = p2g.Fixed[200](addr=2601)
# 2801 .. 2999 - ..
ALARM = p2g.Fixed(addr=3000)
T_MS = p2g.Fixed(addr=3001)
T_HR = p2g.Fixed(addr=3002)
SINGLE_BLOCK_OFF = p2g.Fixed(addr=3003)
FEED_HOLD_OFF = p2g.Fixed(addr=3004)
# 3005 .. 3005 - ..
MESSAGE = p2g.Fixed(addr=3006)
# 3007 .. 3010 - ..
YEAR_MONTH_DAY = p2g.Fixed(addr=3011)
HOUR_MINUTE_SECOND = p2g.Fixed(addr=3012)
# 3013 .. 3019 - ..
POWER_ON_TIME = p2g.Fixed(addr=3020)
CYCLE_START_TIME = p2g.Fixed(addr=3021)
FEED_TIMER = p2g.Fixed(addr=3022)
CUR_PART_TIMER = p2g.Fixed(addr=3023)
LAST_COMPLETE_PART_TIMER = p2g.Fixed(addr=3024)
LAST_PART_TIMER = p2g.Fixed(addr=3025)
TOOL_IN_SPIDLE = p2g.Fixed(addr=3026)
SPINDLE_RPM = p2g.Fixed(addr=3027)
PALLET_LOADED = p2g.Fixed(addr=3028)
# 3029 .. 3029 - ..
SINGLE_BLOCK = p2g.Fixed(addr=3030)
AGAP = p2g.Fixed(addr=3031)
BLOCK_DELETE = p2g.Fixed(addr=3032)
OPT_STOP = p2g.Fixed(addr=3033)
# 3034 .. 3195 - ..
TIMER_CELL_SAFE = p2g.Fixed(addr=3196)
# 3197 .. 3200 - ..
TOOL_TBL_DIAMETER = p2g.Fixed[200](addr=3201)
TOOL_TBL_COOLANT_POSITION = p2g.Fixed[200](addr=3401)
# 3601 .. 3900 - ..
M30_COUNT1 = p2g.Fixed(addr=3901)
M30_COUNT2 = p2g.Fixed(addr=3902)
# 3903 .. 4000 - ..
LAST_BLOCK_G = p2g.Fixed[21](addr=4001)
# 4022 .. 4100 - ..
LAST_BLOCK_ADDRESS = p2g.Fixed[26](addr=4101)
# 4127 .. 5000 - ..
LAST_TARGET_POS = p2g.Fixed[NAXES](addr=5001)
MACHINE_POS = p2g.Fixed[NAXES](addr=5021)
WORK_POS = p2g.Fixed[NAXES](addr=5041)
SKIP_POS = p2g.Fixed[NAXES](addr=5061)
TOOL_OFFSET = p2g.Fixed[20](addr=5081)
# 5101 .. 5200 - ..
G52 = p2g.Fixed[NAXES](addr=5201)
G54 = p2g.Fixed[NAXES](addr=5221)
G55 = p2g.Fixed[NAXES](addr=5241)
G56 = p2g.Fixed[NAXES](addr=5261)
G57 = p2g.Fixed[NAXES](addr=5281)
G58 = p2g.Fixed[NAXES](addr=5301)
G59 = p2g.Fixed[NAXES](addr=5321)
# 5341 .. 5400 - ..
TOOL_TBL_FEED_TIMERS = p2g.Fixed[100](addr=5401)
TOOL_TBL_TOTAL_TIMERS = p2g.Fixed[100](addr=5501)
TOOL_TBL_LIFE_LIMITS = p2g.Fixed[100](addr=5601)
TOOL_TBL_LIFE_COUNTERS = p2g.Fixed[100](addr=5701)
TOOL_TBL_LIFE_MAX_LOADS = p2g.Fixed[100](addr=5801)
TOOL_TBL_LIFE_LOAD_LIMITS = p2g.Fixed[100](addr=5901)
# 6001 .. 6197 - ..
NGC_CF = p2g.Fixed(addr=6198)
# 6199 .. 7000 - ..
G154_P1 = p2g.Fixed[NAXES](addr=7001)
G154_P2 = p2g.Fixed[NAXES](addr=7021)
G154_P3 = p2g.Fixed[NAXES](addr=7041)
G154_P4 = p2g.Fixed[NAXES](addr=7061)
G154_P5 = p2g.Fixed[NAXES](addr=7081)
G154_P6 = p2g.Fixed[NAXES](addr=7101)
G154_P7 = p2g.Fixed[NAXES](addr=7121)
G154_P8 = p2g.Fixed[NAXES](addr=7141)
G154_P9 = p2g.Fixed[NAXES](addr=7161)
G154_P10 = p2g.Fixed[NAXES](addr=7181)
G154_P11 = p2g.Fixed[NAXES](addr=7201)
G154_P12 = p2g.Fixed[NAXES](addr=7221)
G154_P13 = p2g.Fixed[NAXES](addr=7241)
G154_P14 = p2g.Fixed[NAXES](addr=7261)
G154_P15 = p2g.Fixed[NAXES](addr=7281)
G154_P16 = p2g.Fixed[NAXES](addr=7301)
G154_P17 = p2g.Fixed[NAXES](addr=7321)
G154_P18 = p2g.Fixed[NAXES](addr=7341)
G154_P19 = p2g.Fixed[NAXES](addr=7361)
G154_P20 = p2g.Fixed[NAXES](addr=7381)
# 7401 .. 7500 - ..
# 7501 .. 7600 PALLET_PRIORITY ..
# 7601 .. 7700 PALLET_STATUS ..
# 7701 .. 7800 PALLET_PROGRAM ..
# 7801 .. 7900 PALLET_USAGE ..
# 7901 .. 8499 - ..
ATM_ID = p2g.Fixed(addr=8500)
ATM_PERCENT = p2g.Fixed(addr=8501)
ATM_TOTAL_AVL_USAGE = p2g.Fixed(addr=8502)
ATM_TOTAL_AVL_HOLE_COUNT = p2g.Fixed(addr=8503)
ATM_TOTAL_AVL_FEED_TIME = p2g.Fixed(addr=8504)
ATM_TOTAL_AVL_TOTAL_TIME = p2g.Fixed(addr=8505)
# 8506 .. 8509 - ..
ATM_NEXT_TOOL_NUMBER = p2g.Fixed(addr=8510)
ATM_NEXT_TOOL_LIFE = p2g.Fixed(addr=8511)
ATM_NEXT_TOOL_AVL_USAGE = p2g.Fixed(addr=8512)
ATM_NEXT_TOOL_HOLE_COUNT = p2g.Fixed(addr=8513)
ATM_NEXT_TOOL_FEED_TIME = p2g.Fixed(addr=8514)
ATM_NEXT_TOOL_TOTAL_TIME = p2g.Fixed(addr=8515)
# 8516 .. 8549 - ..
TOOL_ID = p2g.Fixed(addr=8550)
TOOL_FLUTES = p2g.Fixed(addr=8551)
TOOL_MAX_VIBRATION = p2g.Fixed(addr=8552)
TOOL_LENGTH_OFFSETS = p2g.Fixed(addr=8553)
TOOL_LENGTH_WEAR = p2g.Fixed(addr=8554)
TOOL_DIAMETER_OFFSETS = p2g.Fixed(addr=8555)
TOOL_DIAMETER_WEAR = p2g.Fixed(addr=8556)
TOOL_ACTUAL_DIAMETER = p2g.Fixed(addr=8557)
TOOL_COOLANT_POSITION = p2g.Fixed(addr=8558)
TOOL_FEED_TIMER = p2g.Fixed(addr=8559)
TOOL_TOTAL_TIMER = p2g.Fixed(addr=8560)
TOOL_LIFE_LIMIT = p2g.Fixed(addr=8561)
TOOL_LIFE_COUNTER = p2g.Fixed(addr=8562)
TOOL_LIFE_MAX_LOAD = p2g.Fixed(addr=8563)
TOOL_LIFE_LOAD_LIMIT = p2g.Fixed(addr=8564)
# 8565 .. 8999 - ..
THERMAL_COMP_ACC = p2g.Fixed(addr=9000)
# 9001 .. 9015 - ..
THERMAL_SPINDLE_COMP_ACC = p2g.Fixed(addr=9016)
# 9017 .. 9999 - ..
GVARIABLES3 = p2g.Fixed[1000](addr=10000)
INPUTS1 = p2g.Fixed[256](addr=11000)
# 11256 .. 11999 - ..
OUTPUT1 = p2g.Fixed[256](addr=12000)
# 12256 .. 12999 - ..
FILTERED_ANALOG1 = p2g.Fixed[13](addr=13000)
COOLANT_LEVEL = p2g.Fixed(addr=13013)
FILTERED_ANALOG2 = p2g.Fixed[50](addr=13014)
# 13064 .. 13999 - ..
SETTING = p2g.Fixed[10000](addr=20000)
PARAMETER = p2g.Fixed[10000](addr=30000)
TOOL_TYP = p2g.Fixed[200](addr=50001)
TOOL_MATERIAL = p2g.Fixed[200](addr=50201)
# 50401 .. 101000 - ..
# 51001 .. 102300 - ..
CURRENT_OFFSET = p2g.Fixed[200](addr=50601)
CURRENT_OFFSET2 = p2g.Fixed[200](addr=50801)
VPS_TEMPLATE_OFFSET = p2g.Fixed[100](addr=51301)
WORK_MATERIAL = p2g.Fixed[200](addr=51401)
VPS_FEEDRATE = p2g.Fixed[200](addr=51601)
APPROX_LENGTH = p2g.Fixed[200](addr=51801)
APPROX_DIAMETER = p2g.Fixed[200](addr=52001)
EDGE_MEASURE_HEIGHT = p2g.Fixed[200](addr=52201)
TOOL_TOLERANCE = p2g.Fixed[200](addr=52401)
PROBE_TYPE = p2g.Fixed[200](addr=52601)
PROBE = p2g.alias(SKIP_POS)
WORK = p2g.alias(WORK_POS)
MACHINE = p2g.alias(MACHINE_POS)
G53 = p2g.alias(MACHINE_POS)
# MACHINE GEN ABOVE



import ast
import builtins
import contextlib
import dataclasses
import typing

from typing import Any

from p2g import err
from p2g import nd
from p2g import stat
from p2g import symbol


@dataclasses.dataclass
class VarScopeSentinel:
    name: str


UNDEF = VarScopeSentinel("nv")
GLOBAL = VarScopeSentinel("gbl")
NONLOCAL = VarScopeSentinel("nonlocal")


@dataclasses.dataclass
class ANamespace:
    parent: "ANamespace"
    guts: typing.Dict

    def __init__(self):
        self.guts = {}
        self.parent = typing.cast(ANamespace, None)

    def get(self, key):
        return self.guts.get(key, UNDEF)

    def __getitem__(self, key):
        return self.guts[key]

    def __setitem__(self, key, value):
        self.guts[key] = value

    def __delitem__(self, key):
        del self.guts[key]

    def __contains__(self, key):
        return key in self.guts


class ModuleNS(ANamespace):
    def __repr__(self):
        return "MODULE"  # for debug


class FunctionNS(ANamespace):
    pass


class ClassNS(ANamespace):
    pass


# return namespace and all the parents, filterint out
# parents which are classes.


def scan_namespaces(ns):
    yield ns
    ns = ns.parent
    while ns:
        if not isinstance(ns, ClassNS):
            yield ns

        ns = ns.parent


def resolve_nonlocal(name, ns):
    while ns:
        res = ns.get(name)
        if res is not UNDEF and res is not NONLOCAL:
            if isinstance(ns, ModuleNS):
                break
            return ns
        ns = ns.parent

    err.compiler(f"No binding for nonlocal '{name}'.")


def lookup_to_ns(res, self, nid):
    if res is GLOBAL:
        dstns = self.module_ns
    elif res is NONLOCAL:
        dstns = resolve_nonlocal(nid, self.ns.parent)
    else:
        dstns = self.ns
    return dstns


def find_ns(first_ns, nid):
    for ns in scan_namespaces(first_ns):
        res = ns.get(nid)
        if res is UNDEF:
            continue
        return res, ns
    return UNDEF, None


def _handle_visit_name_load(self, node):
    res, ns = find_ns(self.ns, node.id)
    if ns is None:
        try:
            return getattr(builtins, node.id)
        except AttributeError:
            err.compiler(f"Name '{node.id}' is not defined.")

    if res is GLOBAL:
        res = self.module_ns.get(node.id)
    elif res is NONLOCAL:
        ns = resolve_nonlocal(node.id, ns.parent)
        res = ns[node.id]
    assert res is not UNDEF
    symbol.Table.remember_load(node.id, res)
    return res


def handle_visit_name_del(self, node):
    res = self.ns.get(node.id)
    if res is UNDEF:
        err.compiler(f"Name '{node.id}' is not defined.")

    dstns = lookup_to_ns(res, self, node.id)
    del dstns[node.id]


def handle_visit_name_store(self, node, store_val):
    res = self.ns.get(node.id)

    if isinstance(res, nd.EBase):
        # variable is already in use, instead of
        # overwriting it,  treat as if writing to
        # <foo>.var
        #        breakpoint()
        setattr(res, "var", store_val)
        return

    dstns = lookup_to_ns(res, self, node.id)

    symbol.Table.remember_store(node.id, store_val)
    dstns[node.id] = store_val


def handle_visit_name(self, node):
    match node.ctx:
        case ast.Del():
            handle_visit_name_del(self, node)
        # in augmented assign, name arrives as with load ctx, but
        # must also be stored there, so catch both.
        case _:
            return _handle_visit_name_load(self, node)
    return None


class LoopContext:
    lcontinue: stat.Label
    lbreak: stat.Label
    lorelse: stat.Label

    prev: "LoopContext"

    def __init__(self, prev):
        self.prev = prev
        self.lcontinue = stat.next_label()
        self.lbreak = stat.next_label()
        self.lorelse = stat.next_label()


class WalkBase:
    ns: "ANamespace"
    module_ns: "ModuleNS"
    file_name: str
    func_name: str

    def __init__(self):
        self.file_name = ""
        self.func_name = ""

        self.module_ns = ModuleNS()
        self.ns = self.module_ns
        self.loop = typing.cast(LoopContext, None)
        self.module_ns = typing.cast(ModuleNS, None)

    # def call(self, *_args, **_kwargs):
    #     raise NotImplementedError

    def visit_slist(self, lst):
        res = None
        for el in lst:
            res = self.visit(el)
        return res

    def update_lastplace(self, node):
        pos = err.NodePlace(
            node.col_offset,
            node.end_col_offset,
            node.lineno,
            self.func_name,
            self.file_name,
        )
        err.mark_pos(pos)

    def visit_fail(self, node, *_):
        name = node.__class__.__name__
        err.compiler(f"Feature '{name.lower()}' not implemented.")

    def _visit_return(self, node):
        if not isinstance(self.ns, FunctionNS):
            err.compiler("'return' outside function.")
        return self.visit(node.value)

    def visit(self, node) -> Any:
        if node is None:
            return None
        self.update_lastplace(node)

        method = "_visit_" + node.__class__.__name__.lower()
        visitor: typing.Callable[[ast.AST], Any] = getattr(self, method, self.visit_fail)
        return visitor(node)

    def visit_store(self, node, store_val) -> None:
        self.update_lastplace(node)

        method = "_visit_store_" + node.__class__.__name__.lower()
        visitor: typing.Callable[[ast.AST, ast.AST], None] = getattr(
            self, method, self.visit_fail
        )
        visitor(node, store_val)

    def _visit_expr(self, node):
        self.visit(node.value)

    @contextlib.contextmanager
    def pushpopns(self, ns):
        ns.parent = self.ns
        self.ns = ns
        yield
        self.ns = self.ns.parent


class WalkNS(WalkBase):
    def _visit_global(self, node):
        for n in node.names:
            if n in self.ns and self.ns[n] is not GLOBAL:  # hard to test right
                err.compiler(f"Name '{n}' before global.")
            # Don't store GLOBAL in the top-level namespace
            if self.ns.parent:
                self.ns[n] = GLOBAL

    def _visit_nonlocal(self, node):
        if isinstance(self.ns, ModuleNS):  # hard to test right
            err.compiler("'nonlocal' declaration not allowed at module level.")
        for n in node.names:
            self.ns[n] = NONLOCAL



import collections
import typing


# symbol tables made by watching every assignment from a vector
# constructor to a name and remembering all the address constants
# used.  walk through all address constants, find the symbols
# assocaited and mark them as useful.  then print them out.
class Table:
    print = False
    print_all = False

    # store in this way to keep multiple definitions.
    name_to_thing: typing.Dict[str, typing.Set] = collections.defaultdict(set)
    addrs_used: typing.Set[int] = set()
    # Called for every store to a name, so remember when a vector is
    # given an association.

    @classmethod
    def remember_store(cls, key, thing):
        cls.remember_load(key, thing)

    @classmethod
    def remember_load(cls, key, thing):
        if key in ("machine", "probe", "relative", "work", "goto", "r9810"):
            return
        try:
            cls.name_to_thing[key].add(thing)
        except TypeError:
            # some things are unhashable.
            pass

    @classmethod
    def add_varref(cls, addr, _pos):
        cls.addrs_used.add(int(addr))

    @classmethod
    def yield_lines(cls):
        #        breakpoint()
        if not cls.print and not cls.print_all:
            return

        # get object names and sort them.
        sorted_names = sorted(cls.name_to_thing.keys(), key=str.casefold)

        # go through table of all known macro names,
        # find out if used, and print nicely.
        def yield_table_part(phase):
            old_rhs = ""
            for key in sorted_names:
                new_values = cls.name_to_thing[key]
                for rhsobj in new_values:
                    try:
                        if not cls.print_all and not rhsobj.user_defined:
                            continue
                        new_rhs = rhsobj.to_symtab_entry(cls.addrs_used)
                        if phase in new_rhs:
                            if new_rhs != old_rhs:
                                yield (key, new_rhs)
                                old_rhs = new_rhs
                    except AttributeError:
                        continue

        def yield_all_parts():
            for phase in ["#", ",", "xyz"]:
                yield from yield_table_part(phase)

        lcols, rcols = zip(*yield_all_parts())

        def max_str_len(lines):
            return len(max(lines, key=len, default=""))

        lsize = max_str_len(lcols)
        rsize = max_str_len(rcols)
        for key, value in zip(lcols, rcols):
            yield "( " + key.ljust(lsize) + " : " + value.ljust(rsize) + " )"

    @classmethod
    def reset(cls):
        cls.name_to_thing = collections.defaultdict(set)
        cls.addrs_used = set()
        cls.print = False
